// Code generated by protoc-gen-gogo. DO NOT EDIT.
// source: node.proto

package rangevectorsplitting

import (
	fmt "fmt"
	_ "github.com/gogo/protobuf/gogoproto"
	proto "github.com/gogo/protobuf/proto"
	_ "google.golang.org/protobuf/types/known/durationpb"
	io "io"
	math "math"
	math_bits "math/bits"
	reflect "reflect"
	strings "strings"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.GoGoProtoPackageIsVersion3 // please upgrade the proto package

type SplitFunctionCallDetails struct {
	SplitRanges       []SplitRange `protobuf:"bytes,1,rep,name=splitRanges,proto3" json:"splitRanges"`
	InnerNodeCacheKey string       `protobuf:"bytes,2,opt,name=innerNodeCacheKey,proto3" json:"innerNodeCacheKey,omitempty"`
}

func (m *SplitFunctionCallDetails) Reset()      { *m = SplitFunctionCallDetails{} }
func (*SplitFunctionCallDetails) ProtoMessage() {}
func (*SplitFunctionCallDetails) Descriptor() ([]byte, []int) {
	return fileDescriptor_0c843d59d2d938e7, []int{0}
}
func (m *SplitFunctionCallDetails) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *SplitFunctionCallDetails) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_SplitFunctionCallDetails.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *SplitFunctionCallDetails) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SplitFunctionCallDetails.Merge(m, src)
}
func (m *SplitFunctionCallDetails) XXX_Size() int {
	return m.Size()
}
func (m *SplitFunctionCallDetails) XXX_DiscardUnknown() {
	xxx_messageInfo_SplitFunctionCallDetails.DiscardUnknown(m)
}

var xxx_messageInfo_SplitFunctionCallDetails proto.InternalMessageInfo

func (m *SplitFunctionCallDetails) GetSplitRanges() []SplitRange {
	if m != nil {
		return m.SplitRanges
	}
	return nil
}

func (m *SplitFunctionCallDetails) GetInnerNodeCacheKey() string {
	if m != nil {
		return m.InnerNodeCacheKey
	}
	return ""
}

func (*SplitFunctionCallDetails) XXX_MessageName() string {
	return "rangevectorsplitting.SplitFunctionCallDetails"
}

// SplitRange represents a time range within a split.
// Start is exclusive (points with timestamp > Start are included).
// End is inclusive (points with timestamp <= End are included).
type SplitRange struct {
	Start     int64 `protobuf:"varint,1,opt,name=start,proto3" json:"start,omitempty"`
	End       int64 `protobuf:"varint,2,opt,name=end,proto3" json:"end,omitempty"`
	Cacheable bool  `protobuf:"varint,3,opt,name=cacheable,proto3" json:"cacheable,omitempty"`
}

func (m *SplitRange) Reset()      { *m = SplitRange{} }
func (*SplitRange) ProtoMessage() {}
func (*SplitRange) Descriptor() ([]byte, []int) {
	return fileDescriptor_0c843d59d2d938e7, []int{1}
}
func (m *SplitRange) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *SplitRange) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_SplitRange.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *SplitRange) XXX_Merge(src proto.Message) {
	xxx_messageInfo_SplitRange.Merge(m, src)
}
func (m *SplitRange) XXX_Size() int {
	return m.Size()
}
func (m *SplitRange) XXX_DiscardUnknown() {
	xxx_messageInfo_SplitRange.DiscardUnknown(m)
}

var xxx_messageInfo_SplitRange proto.InternalMessageInfo

func (m *SplitRange) GetStart() int64 {
	if m != nil {
		return m.Start
	}
	return 0
}

func (m *SplitRange) GetEnd() int64 {
	if m != nil {
		return m.End
	}
	return 0
}

func (m *SplitRange) GetCacheable() bool {
	if m != nil {
		return m.Cacheable
	}
	return false
}

func (*SplitRange) XXX_MessageName() string {
	return "rangevectorsplitting.SplitRange"
}
func init() {
	proto.RegisterType((*SplitFunctionCallDetails)(nil), "rangevectorsplitting.SplitFunctionCallDetails")
	proto.RegisterType((*SplitRange)(nil), "rangevectorsplitting.SplitRange")
}

func init() { proto.RegisterFile("node.proto", fileDescriptor_0c843d59d2d938e7) }

var fileDescriptor_0c843d59d2d938e7 = []byte{
	// 299 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0x6c, 0x90, 0xbf, 0x4a, 0x03, 0x31,
	0x1c, 0xc7, 0xf3, 0xf3, 0x54, 0x6c, 0xba, 0xe8, 0xd1, 0xe1, 0x28, 0xf2, 0xf3, 0xe8, 0x74, 0x83,
	0xb4, 0xa0, 0x8b, 0x73, 0x2b, 0x22, 0x08, 0x0e, 0xe9, 0x13, 0xa4, 0x77, 0x31, 0x1e, 0x84, 0xa4,
	0xe4, 0x52, 0xc1, 0xcd, 0x47, 0x10, 0x9f, 0xc2, 0x47, 0xe9, 0xd8, 0xb1, 0x93, 0xd8, 0xdc, 0xe2,
	0xd8, 0x47, 0x90, 0xdc, 0x81, 0x15, 0x74, 0xfb, 0x7d, 0xff, 0xe4, 0xf3, 0x85, 0x50, 0xaa, 0x4d,
	0x21, 0x86, 0x73, 0x6b, 0x9c, 0x89, 0x7b, 0x96, 0x6b, 0x29, 0x9e, 0x44, 0xee, 0x8c, 0xad, 0xe6,
	0xaa, 0x74, 0xae, 0xd4, 0xb2, 0xdf, 0x93, 0x46, 0x9a, 0xa6, 0x30, 0x0a, 0x57, 0xdb, 0xed, 0xa3,
	0x34, 0x46, 0x2a, 0x31, 0x6a, 0xd4, 0x6c, 0xf1, 0x30, 0x2a, 0x16, 0x96, 0xbb, 0xd2, 0xe8, 0x36,
	0x1f, 0xbc, 0x01, 0x4d, 0xa6, 0x81, 0x71, 0xb3, 0xd0, 0x79, 0xf0, 0x27, 0x5c, 0xa9, 0x6b, 0xe1,
	0x78, 0xa9, 0xaa, 0xf8, 0x96, 0x76, 0x1b, 0x3e, 0x0b, 0x7b, 0x55, 0x02, 0x69, 0x94, 0x75, 0x2f,
	0xd2, 0xe1, 0x7f, 0xf3, 0xc3, 0xe9, 0x4f, 0x71, 0xbc, 0xbf, 0xfc, 0x38, 0x23, 0xec, 0xf7, 0xd3,
	0xf8, 0x9c, 0x9e, 0x94, 0x5a, 0x0b, 0x7b, 0x6f, 0x0a, 0x31, 0xe1, 0xf9, 0xa3, 0xb8, 0x13, 0xcf,
	0xc9, 0x5e, 0x0a, 0x59, 0x87, 0xfd, 0x0d, 0x06, 0x8c, 0xd2, 0x1d, 0x2e, 0xee, 0xd1, 0x83, 0xca,
	0x71, 0xeb, 0x12, 0x48, 0x21, 0x8b, 0x58, 0x2b, 0xe2, 0x63, 0x1a, 0x09, 0x5d, 0x34, 0x8c, 0x88,
	0x85, 0x33, 0x3e, 0xa5, 0x9d, 0x3c, 0x10, 0xf8, 0x4c, 0x89, 0x24, 0x4a, 0x21, 0x3b, 0x62, 0x3b,
	0x63, 0x7c, 0xb5, 0xda, 0x20, 0x59, 0x6f, 0x90, 0x6c, 0x37, 0x08, 0x2f, 0x1e, 0xe1, 0xdd, 0x23,
	0x59, 0x7a, 0x84, 0x95, 0x47, 0xf8, 0xf4, 0x08, 0x5f, 0x1e, 0xc9, 0xd6, 0x23, 0xbc, 0xd6, 0x48,
	0x96, 0x35, 0xc2, 0xaa, 0x46, 0xb2, 0xae, 0x91, 0xcc, 0x0e, 0x9b, 0x9f, 0xba, 0xfc, 0x0e, 0x00,
	0x00, 0xff, 0xff, 0xe2, 0x1e, 0xdb, 0x07, 0x83, 0x01, 0x00, 0x00,
}

func (this *SplitFunctionCallDetails) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 6)
	s = append(s, "&rangevectorsplitting.SplitFunctionCallDetails{")
	if this.SplitRanges != nil {
		vs := make([]SplitRange, len(this.SplitRanges))
		for i := range vs {
			vs[i] = this.SplitRanges[i]
		}
		s = append(s, "SplitRanges: "+fmt.Sprintf("%#v", vs)+",\n")
	}
	s = append(s, "InnerNodeCacheKey: "+fmt.Sprintf("%#v", this.InnerNodeCacheKey)+",\n")
	s = append(s, "}")
	return strings.Join(s, "")
}
func (this *SplitRange) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 7)
	s = append(s, "&rangevectorsplitting.SplitRange{")
	s = append(s, "Start: "+fmt.Sprintf("%#v", this.Start)+",\n")
	s = append(s, "End: "+fmt.Sprintf("%#v", this.End)+",\n")
	s = append(s, "Cacheable: "+fmt.Sprintf("%#v", this.Cacheable)+",\n")
	s = append(s, "}")
	return strings.Join(s, "")
}
func valueToGoStringNode(v interface{}, typ string) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("func(v %v) *%v { return &v } ( %#v )", typ, typ, pv)
}
func (m *SplitFunctionCallDetails) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *SplitFunctionCallDetails) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *SplitFunctionCallDetails) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if len(m.InnerNodeCacheKey) > 0 {
		i -= len(m.InnerNodeCacheKey)
		copy(dAtA[i:], m.InnerNodeCacheKey)
		i = encodeVarintNode(dAtA, i, uint64(len(m.InnerNodeCacheKey)))
		i--
		dAtA[i] = 0x12
	}
	if len(m.SplitRanges) > 0 {
		for iNdEx := len(m.SplitRanges) - 1; iNdEx >= 0; iNdEx-- {
			{
				size, err := m.SplitRanges[iNdEx].MarshalToSizedBuffer(dAtA[:i])
				if err != nil {
					return 0, err
				}
				i -= size
				i = encodeVarintNode(dAtA, i, uint64(size))
			}
			i--
			dAtA[i] = 0xa
		}
	}
	return len(dAtA) - i, nil
}

func (m *SplitRange) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *SplitRange) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *SplitRange) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if m.Cacheable {
		i--
		if m.Cacheable {
			dAtA[i] = 1
		} else {
			dAtA[i] = 0
		}
		i--
		dAtA[i] = 0x18
	}
	if m.End != 0 {
		i = encodeVarintNode(dAtA, i, uint64(m.End))
		i--
		dAtA[i] = 0x10
	}
	if m.Start != 0 {
		i = encodeVarintNode(dAtA, i, uint64(m.Start))
		i--
		dAtA[i] = 0x8
	}
	return len(dAtA) - i, nil
}

func encodeVarintNode(dAtA []byte, offset int, v uint64) int {
	offset -= sovNode(v)
	base := offset
	for v >= 1<<7 {
		dAtA[offset] = uint8(v&0x7f | 0x80)
		v >>= 7
		offset++
	}
	dAtA[offset] = uint8(v)
	return base
}
func (m *SplitFunctionCallDetails) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if len(m.SplitRanges) > 0 {
		for _, e := range m.SplitRanges {
			l = e.Size()
			n += 1 + l + sovNode(uint64(l))
		}
	}
	l = len(m.InnerNodeCacheKey)
	if l > 0 {
		n += 1 + l + sovNode(uint64(l))
	}
	return n
}

func (m *SplitRange) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if m.Start != 0 {
		n += 1 + sovNode(uint64(m.Start))
	}
	if m.End != 0 {
		n += 1 + sovNode(uint64(m.End))
	}
	if m.Cacheable {
		n += 2
	}
	return n
}

func sovNode(x uint64) (n int) {
	return (math_bits.Len64(x|1) + 6) / 7
}
func sozNode(x uint64) (n int) {
	return sovNode(uint64((x << 1) ^ uint64((int64(x) >> 63))))
}
func (this *SplitFunctionCallDetails) String() string {
	if this == nil {
		return "nil"
	}
	repeatedStringForSplitRanges := "[]SplitRange{"
	for _, f := range this.SplitRanges {
		repeatedStringForSplitRanges += strings.Replace(strings.Replace(f.String(), "SplitRange", "SplitRange", 1), `&`, ``, 1) + ","
	}
	repeatedStringForSplitRanges += "}"
	s := strings.Join([]string{`&SplitFunctionCallDetails{`,
		`SplitRanges:` + repeatedStringForSplitRanges + `,`,
		`InnerNodeCacheKey:` + fmt.Sprintf("%v", this.InnerNodeCacheKey) + `,`,
		`}`,
	}, "")
	return s
}
func (this *SplitRange) String() string {
	if this == nil {
		return "nil"
	}
	s := strings.Join([]string{`&SplitRange{`,
		`Start:` + fmt.Sprintf("%v", this.Start) + `,`,
		`End:` + fmt.Sprintf("%v", this.End) + `,`,
		`Cacheable:` + fmt.Sprintf("%v", this.Cacheable) + `,`,
		`}`,
	}, "")
	return s
}
func valueToStringNode(v interface{}) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("*%v", pv)
}
func (m *SplitFunctionCallDetails) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowNode
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: SplitFunctionCallDetails: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: SplitFunctionCallDetails: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 2 {
				return fmt.Errorf("proto: wrong wireType = %d for field SplitRanges", wireType)
			}
			var msglen int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				msglen |= int(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			if msglen < 0 {
				return ErrInvalidLengthNode
			}
			postIndex := iNdEx + msglen
			if postIndex < 0 {
				return ErrInvalidLengthNode
			}
			if postIndex > l {
				return io.ErrUnexpectedEOF
			}
			m.SplitRanges = append(m.SplitRanges, SplitRange{})
			if err := m.SplitRanges[len(m.SplitRanges)-1].Unmarshal(dAtA[iNdEx:postIndex]); err != nil {
				return err
			}
			iNdEx = postIndex
		case 2:
			if wireType != 2 {
				return fmt.Errorf("proto: wrong wireType = %d for field InnerNodeCacheKey", wireType)
			}
			var stringLen uint64
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				stringLen |= uint64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			intStringLen := int(stringLen)
			if intStringLen < 0 {
				return ErrInvalidLengthNode
			}
			postIndex := iNdEx + intStringLen
			if postIndex < 0 {
				return ErrInvalidLengthNode
			}
			if postIndex > l {
				return io.ErrUnexpectedEOF
			}
			m.InnerNodeCacheKey = string(dAtA[iNdEx:postIndex])
			iNdEx = postIndex
		default:
			iNdEx = preIndex
			skippy, err := skipNode(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if (skippy < 0) || (iNdEx+skippy) < 0 {
				return ErrInvalidLengthNode
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func (m *SplitRange) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowNode
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: SplitRange: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: SplitRange: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field Start", wireType)
			}
			m.Start = 0
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				m.Start |= int64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
		case 2:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field End", wireType)
			}
			m.End = 0
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				m.End |= int64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
		case 3:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field Cacheable", wireType)
			}
			var v int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				v |= int(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			m.Cacheable = bool(v != 0)
		default:
			iNdEx = preIndex
			skippy, err := skipNode(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if (skippy < 0) || (iNdEx+skippy) < 0 {
				return ErrInvalidLengthNode
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func skipNode(dAtA []byte) (n int, err error) {
	l := len(dAtA)
	iNdEx := 0
	depth := 0
	for iNdEx < l {
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return 0, ErrIntOverflowNode
			}
			if iNdEx >= l {
				return 0, io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= (uint64(b) & 0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		wireType := int(wire & 0x7)
		switch wireType {
		case 0:
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowNode
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				iNdEx++
				if dAtA[iNdEx-1] < 0x80 {
					break
				}
			}
		case 1:
			iNdEx += 8
		case 2:
			var length int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowNode
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				length |= (int(b) & 0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			if length < 0 {
				return 0, ErrInvalidLengthNode
			}
			iNdEx += length
		case 3:
			depth++
		case 4:
			if depth == 0 {
				return 0, ErrUnexpectedEndOfGroupNode
			}
			depth--
		case 5:
			iNdEx += 4
		default:
			return 0, fmt.Errorf("proto: illegal wireType %d", wireType)
		}
		if iNdEx < 0 {
			return 0, ErrInvalidLengthNode
		}
		if depth == 0 {
			return iNdEx, nil
		}
	}
	return 0, io.ErrUnexpectedEOF
}

var (
	ErrInvalidLengthNode        = fmt.Errorf("proto: negative length found during unmarshaling")
	ErrIntOverflowNode          = fmt.Errorf("proto: integer overflow")
	ErrUnexpectedEndOfGroupNode = fmt.Errorf("proto: unexpected end of group")
)
