// Code generated by protoc-gen-gogo. DO NOT EDIT.
// source: propagation_tracker.proto

package memberlist

import (
	fmt "fmt"
	_ "github.com/gogo/protobuf/gogoproto"
	proto "github.com/gogo/protobuf/proto"
	github_com_gogo_protobuf_sortkeys "github.com/gogo/protobuf/sortkeys"
	io "io"
	math "math"
	math_bits "math/bits"
	reflect "reflect"
	strings "strings"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.GoGoProtoPackageIsVersion3 // please upgrade the proto package

// PropagationDelayTrackerDesc stores beacons for tracking gossip propagation delay
// across the cluster. Beacons are cluster-wide (not per-node) and are keyed
// by a random beacon ID.
type PropagationDelayTrackerDesc struct {
	// Map of beacon_id -> beacon.
	Beacons map[uint64]BeaconDesc `protobuf:"bytes,1,rep,name=beacons,proto3" json:"beacons" protobuf_key:"varint,1,opt,name=key,proto3" protobuf_val:"bytes,2,opt,name=value,proto3"`
}

func (m *PropagationDelayTrackerDesc) Reset()      { *m = PropagationDelayTrackerDesc{} }
func (*PropagationDelayTrackerDesc) ProtoMessage() {}
func (*PropagationDelayTrackerDesc) Descriptor() ([]byte, []int) {
	return fileDescriptor_cb1c69f252ca5e51, []int{0}
}
func (m *PropagationDelayTrackerDesc) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *PropagationDelayTrackerDesc) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_PropagationDelayTrackerDesc.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *PropagationDelayTrackerDesc) XXX_Merge(src proto.Message) {
	xxx_messageInfo_PropagationDelayTrackerDesc.Merge(m, src)
}
func (m *PropagationDelayTrackerDesc) XXX_Size() int {
	return m.Size()
}
func (m *PropagationDelayTrackerDesc) XXX_DiscardUnknown() {
	xxx_messageInfo_PropagationDelayTrackerDesc.DiscardUnknown(m)
}

var xxx_messageInfo_PropagationDelayTrackerDesc proto.InternalMessageInfo

func (m *PropagationDelayTrackerDesc) GetBeacons() map[uint64]BeaconDesc {
	if m != nil {
		return m.Beacons
	}
	return nil
}

// Beacon represents a single propagation beacon published to the cluster.
type BeaconDesc struct {
	// Name of the node that published this beacon.
	PublishedBy string `protobuf:"bytes,3,opt,name=published_by,json=publishedBy,proto3" json:"published_by,omitempty"`
	// Wall-clock timestamp when this beacon was published (milliseconds since epoch).
	PublishedAt int64 `protobuf:"varint,1,opt,name=published_at,json=publishedAt,proto3" json:"published_at,omitempty"`
	// Deletion timestamp (milliseconds since epoch). 0 = active, non-zero = beacon deleted at that time (tombstone).
	DeletedAt int64 `protobuf:"varint,2,opt,name=deleted_at,json=deletedAt,proto3" json:"deleted_at,omitempty"`
}

func (m *BeaconDesc) Reset()      { *m = BeaconDesc{} }
func (*BeaconDesc) ProtoMessage() {}
func (*BeaconDesc) Descriptor() ([]byte, []int) {
	return fileDescriptor_cb1c69f252ca5e51, []int{1}
}
func (m *BeaconDesc) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *BeaconDesc) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_BeaconDesc.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *BeaconDesc) XXX_Merge(src proto.Message) {
	xxx_messageInfo_BeaconDesc.Merge(m, src)
}
func (m *BeaconDesc) XXX_Size() int {
	return m.Size()
}
func (m *BeaconDesc) XXX_DiscardUnknown() {
	xxx_messageInfo_BeaconDesc.DiscardUnknown(m)
}

var xxx_messageInfo_BeaconDesc proto.InternalMessageInfo

func (m *BeaconDesc) GetPublishedBy() string {
	if m != nil {
		return m.PublishedBy
	}
	return ""
}

func (m *BeaconDesc) GetPublishedAt() int64 {
	if m != nil {
		return m.PublishedAt
	}
	return 0
}

func (m *BeaconDesc) GetDeletedAt() int64 {
	if m != nil {
		return m.DeletedAt
	}
	return 0
}

func init() {
	proto.RegisterType((*PropagationDelayTrackerDesc)(nil), "memberlist.PropagationDelayTrackerDesc")
	proto.RegisterMapType((map[uint64]BeaconDesc)(nil), "memberlist.PropagationDelayTrackerDesc.BeaconsEntry")
	proto.RegisterType((*BeaconDesc)(nil), "memberlist.BeaconDesc")
}

func init() { proto.RegisterFile("propagation_tracker.proto", fileDescriptor_cb1c69f252ca5e51) }

var fileDescriptor_cb1c69f252ca5e51 = []byte{
	// 330 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0x7c, 0x90, 0x31, 0x4f, 0xfa, 0x40,
	0x18, 0xc6, 0xef, 0xa5, 0xfc, 0xff, 0x86, 0x83, 0xc1, 0x74, 0x30, 0x15, 0xe3, 0x59, 0x99, 0x3a,
	0x68, 0x49, 0x90, 0xc1, 0xb8, 0xd1, 0xe0, 0x6e, 0x1a, 0x76, 0x72, 0x57, 0xce, 0xd2, 0x50, 0xb8,
	0xda, 0x5e, 0x4d, 0xba, 0xf9, 0x11, 0xfc, 0x18, 0x7e, 0x13, 0x19, 0x19, 0x99, 0x8c, 0x1c, 0x8b,
	0x23, 0x1f, 0xc1, 0x70, 0x55, 0xc1, 0xc5, 0xed, 0x7d, 0xdf, 0xe7, 0xf7, 0x3c, 0x4f, 0xf2, 0xe2,
	0xe3, 0x24, 0x15, 0x09, 0x0d, 0xa9, 0x8c, 0xc4, 0x6c, 0x28, 0x53, 0x1a, 0x4c, 0x78, 0xea, 0x26,
	0xa9, 0x90, 0xc2, 0xc4, 0x53, 0x3e, 0x65, 0x3c, 0x8d, 0xa3, 0x4c, 0x36, 0x2f, 0xc3, 0x48, 0x8e,
	0x73, 0xe6, 0x06, 0x62, 0xda, 0x0e, 0x45, 0x28, 0xda, 0x1a, 0x61, 0xf9, 0xbd, 0xde, 0xf4, 0xa2,
	0xa7, 0xd2, 0xda, 0x7a, 0x05, 0x7c, 0x72, 0xb7, 0x0b, 0xee, 0xf3, 0x98, 0x16, 0x83, 0x32, 0xbc,
	0xcf, 0xb3, 0xc0, 0x1c, 0xe0, 0x03, 0xc6, 0x69, 0x20, 0x66, 0x99, 0x05, 0xb6, 0xe1, 0xd4, 0x3b,
	0x5d, 0x77, 0x57, 0xe6, 0xfe, 0xe1, 0x74, 0xbd, 0xd2, 0x76, 0x3b, 0x93, 0x69, 0xe1, 0x55, 0xe7,
	0x6f, 0x67, 0xc8, 0xff, 0x8e, 0x6a, 0xfa, 0xb8, 0xb1, 0x2f, 0x9b, 0x87, 0xd8, 0x98, 0xf0, 0xc2,
	0x02, 0x1b, 0x9c, 0xaa, 0xbf, 0x1d, 0xcd, 0x0b, 0xfc, 0xef, 0x91, 0xc6, 0x39, 0xb7, 0x2a, 0x36,
	0x38, 0xf5, 0xce, 0xd1, 0x7e, 0x6b, 0x69, 0xdd, 0x96, 0xf8, 0x25, 0x74, 0x53, 0xb9, 0x86, 0xd6,
	0x03, 0xc6, 0x3b, 0xc1, 0x3c, 0xc7, 0x8d, 0x24, 0x67, 0x71, 0x94, 0x8d, 0xf9, 0x68, 0xc8, 0x0a,
	0xcb, 0xb0, 0xc1, 0xa9, 0xf9, 0xf5, 0x9f, 0x9b, 0x57, 0xfc, 0x46, 0xa8, 0xd4, 0xed, 0xc6, 0x1e,
	0xd2, 0x93, 0xe6, 0x29, 0xc6, 0x23, 0x1e, 0x73, 0x59, 0x02, 0x15, 0x0d, 0xd4, 0xbe, 0x2e, 0x3d,
	0xe9, 0x75, 0x17, 0x2b, 0x82, 0x96, 0x2b, 0x82, 0x36, 0x2b, 0x02, 0x4f, 0x8a, 0xc0, 0x8b, 0x22,
	0x30, 0x57, 0x04, 0x16, 0x8a, 0xc0, 0xbb, 0x22, 0xf0, 0xa1, 0x08, 0xda, 0x28, 0x02, 0xcf, 0x6b,
	0x82, 0x16, 0x6b, 0x82, 0x96, 0x6b, 0x82, 0xd8, 0x7f, 0xfd, 0xf9, 0xab, 0xcf, 0x00, 0x00, 0x00,
	0xff, 0xff, 0xc6, 0xa5, 0x91, 0x8c, 0xd1, 0x01, 0x00, 0x00,
}

func (this *PropagationDelayTrackerDesc) Equal(that interface{}) bool {
	if that == nil {
		return this == nil
	}

	that1, ok := that.(*PropagationDelayTrackerDesc)
	if !ok {
		that2, ok := that.(PropagationDelayTrackerDesc)
		if ok {
			that1 = &that2
		} else {
			return false
		}
	}
	if that1 == nil {
		return this == nil
	} else if this == nil {
		return false
	}
	if len(this.Beacons) != len(that1.Beacons) {
		return false
	}
	for i := range this.Beacons {
		a := this.Beacons[i]
		b := that1.Beacons[i]
		if !(&a).Equal(&b) {
			return false
		}
	}
	return true
}
func (this *BeaconDesc) Equal(that interface{}) bool {
	if that == nil {
		return this == nil
	}

	that1, ok := that.(*BeaconDesc)
	if !ok {
		that2, ok := that.(BeaconDesc)
		if ok {
			that1 = &that2
		} else {
			return false
		}
	}
	if that1 == nil {
		return this == nil
	} else if this == nil {
		return false
	}
	if this.PublishedBy != that1.PublishedBy {
		return false
	}
	if this.PublishedAt != that1.PublishedAt {
		return false
	}
	if this.DeletedAt != that1.DeletedAt {
		return false
	}
	return true
}
func (this *PropagationDelayTrackerDesc) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 5)
	s = append(s, "&memberlist.PropagationDelayTrackerDesc{")
	keysForBeacons := make([]uint64, 0, len(this.Beacons))
	for k, _ := range this.Beacons {
		keysForBeacons = append(keysForBeacons, k)
	}
	github_com_gogo_protobuf_sortkeys.Uint64s(keysForBeacons)
	mapStringForBeacons := "map[uint64]BeaconDesc{"
	for _, k := range keysForBeacons {
		mapStringForBeacons += fmt.Sprintf("%#v: %#v,", k, this.Beacons[k])
	}
	mapStringForBeacons += "}"
	if this.Beacons != nil {
		s = append(s, "Beacons: "+mapStringForBeacons+",\n")
	}
	s = append(s, "}")
	return strings.Join(s, "")
}
func (this *BeaconDesc) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 7)
	s = append(s, "&memberlist.BeaconDesc{")
	s = append(s, "PublishedBy: "+fmt.Sprintf("%#v", this.PublishedBy)+",\n")
	s = append(s, "PublishedAt: "+fmt.Sprintf("%#v", this.PublishedAt)+",\n")
	s = append(s, "DeletedAt: "+fmt.Sprintf("%#v", this.DeletedAt)+",\n")
	s = append(s, "}")
	return strings.Join(s, "")
}
func valueToGoStringPropagationTracker(v interface{}, typ string) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("func(v %v) *%v { return &v } ( %#v )", typ, typ, pv)
}
func (m *PropagationDelayTrackerDesc) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *PropagationDelayTrackerDesc) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *PropagationDelayTrackerDesc) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if len(m.Beacons) > 0 {
		for k := range m.Beacons {
			v := m.Beacons[k]
			baseI := i
			{
				size, err := (&v).MarshalToSizedBuffer(dAtA[:i])
				if err != nil {
					return 0, err
				}
				i -= size
				i = encodeVarintPropagationTracker(dAtA, i, uint64(size))
			}
			i--
			dAtA[i] = 0x12
			i = encodeVarintPropagationTracker(dAtA, i, uint64(k))
			i--
			dAtA[i] = 0x8
			i = encodeVarintPropagationTracker(dAtA, i, uint64(baseI-i))
			i--
			dAtA[i] = 0xa
		}
	}
	return len(dAtA) - i, nil
}

func (m *BeaconDesc) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *BeaconDesc) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *BeaconDesc) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if len(m.PublishedBy) > 0 {
		i -= len(m.PublishedBy)
		copy(dAtA[i:], m.PublishedBy)
		i = encodeVarintPropagationTracker(dAtA, i, uint64(len(m.PublishedBy)))
		i--
		dAtA[i] = 0x1a
	}
	if m.DeletedAt != 0 {
		i = encodeVarintPropagationTracker(dAtA, i, uint64(m.DeletedAt))
		i--
		dAtA[i] = 0x10
	}
	if m.PublishedAt != 0 {
		i = encodeVarintPropagationTracker(dAtA, i, uint64(m.PublishedAt))
		i--
		dAtA[i] = 0x8
	}
	return len(dAtA) - i, nil
}

func encodeVarintPropagationTracker(dAtA []byte, offset int, v uint64) int {
	offset -= sovPropagationTracker(v)
	base := offset
	for v >= 1<<7 {
		dAtA[offset] = uint8(v&0x7f | 0x80)
		v >>= 7
		offset++
	}
	dAtA[offset] = uint8(v)
	return base
}
func (m *PropagationDelayTrackerDesc) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if len(m.Beacons) > 0 {
		for k, v := range m.Beacons {
			_ = k
			_ = v
			l = v.Size()
			mapEntrySize := 1 + sovPropagationTracker(uint64(k)) + 1 + l + sovPropagationTracker(uint64(l))
			n += mapEntrySize + 1 + sovPropagationTracker(uint64(mapEntrySize))
		}
	}
	return n
}

func (m *BeaconDesc) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if m.PublishedAt != 0 {
		n += 1 + sovPropagationTracker(uint64(m.PublishedAt))
	}
	if m.DeletedAt != 0 {
		n += 1 + sovPropagationTracker(uint64(m.DeletedAt))
	}
	l = len(m.PublishedBy)
	if l > 0 {
		n += 1 + l + sovPropagationTracker(uint64(l))
	}
	return n
}

func sovPropagationTracker(x uint64) (n int) {
	return (math_bits.Len64(x|1) + 6) / 7
}
func sozPropagationTracker(x uint64) (n int) {
	return sovPropagationTracker(uint64((x << 1) ^ uint64((int64(x) >> 63))))
}
func (this *PropagationDelayTrackerDesc) String() string {
	if this == nil {
		return "nil"
	}
	keysForBeacons := make([]uint64, 0, len(this.Beacons))
	for k, _ := range this.Beacons {
		keysForBeacons = append(keysForBeacons, k)
	}
	github_com_gogo_protobuf_sortkeys.Uint64s(keysForBeacons)
	mapStringForBeacons := "map[uint64]BeaconDesc{"
	for _, k := range keysForBeacons {
		mapStringForBeacons += fmt.Sprintf("%v: %v,", k, this.Beacons[k])
	}
	mapStringForBeacons += "}"
	s := strings.Join([]string{`&PropagationDelayTrackerDesc{`,
		`Beacons:` + mapStringForBeacons + `,`,
		`}`,
	}, "")
	return s
}
func (this *BeaconDesc) String() string {
	if this == nil {
		return "nil"
	}
	s := strings.Join([]string{`&BeaconDesc{`,
		`PublishedAt:` + fmt.Sprintf("%v", this.PublishedAt) + `,`,
		`DeletedAt:` + fmt.Sprintf("%v", this.DeletedAt) + `,`,
		`PublishedBy:` + fmt.Sprintf("%v", this.PublishedBy) + `,`,
		`}`,
	}, "")
	return s
}
func valueToStringPropagationTracker(v interface{}) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("*%v", pv)
}
func (m *PropagationDelayTrackerDesc) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowPropagationTracker
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: PropagationDelayTrackerDesc: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: PropagationDelayTrackerDesc: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 2 {
				return fmt.Errorf("proto: wrong wireType = %d for field Beacons", wireType)
			}
			var msglen int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				msglen |= int(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			if msglen < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			postIndex := iNdEx + msglen
			if postIndex < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if postIndex > l {
				return io.ErrUnexpectedEOF
			}
			if m.Beacons == nil {
				m.Beacons = make(map[uint64]BeaconDesc)
			}
			var mapkey uint64
			mapvalue := &BeaconDesc{}
			for iNdEx < postIndex {
				entryPreIndex := iNdEx
				var wire uint64
				for shift := uint(0); ; shift += 7 {
					if shift >= 64 {
						return ErrIntOverflowPropagationTracker
					}
					if iNdEx >= l {
						return io.ErrUnexpectedEOF
					}
					b := dAtA[iNdEx]
					iNdEx++
					wire |= uint64(b&0x7F) << shift
					if b < 0x80 {
						break
					}
				}
				fieldNum := int32(wire >> 3)
				if fieldNum == 1 {
					for shift := uint(0); ; shift += 7 {
						if shift >= 64 {
							return ErrIntOverflowPropagationTracker
						}
						if iNdEx >= l {
							return io.ErrUnexpectedEOF
						}
						b := dAtA[iNdEx]
						iNdEx++
						mapkey |= uint64(b&0x7F) << shift
						if b < 0x80 {
							break
						}
					}
				} else if fieldNum == 2 {
					var mapmsglen int
					for shift := uint(0); ; shift += 7 {
						if shift >= 64 {
							return ErrIntOverflowPropagationTracker
						}
						if iNdEx >= l {
							return io.ErrUnexpectedEOF
						}
						b := dAtA[iNdEx]
						iNdEx++
						mapmsglen |= int(b&0x7F) << shift
						if b < 0x80 {
							break
						}
					}
					if mapmsglen < 0 {
						return ErrInvalidLengthPropagationTracker
					}
					postmsgIndex := iNdEx + mapmsglen
					if postmsgIndex < 0 {
						return ErrInvalidLengthPropagationTracker
					}
					if postmsgIndex > l {
						return io.ErrUnexpectedEOF
					}
					mapvalue = &BeaconDesc{}
					if err := mapvalue.Unmarshal(dAtA[iNdEx:postmsgIndex]); err != nil {
						return err
					}
					iNdEx = postmsgIndex
				} else {
					iNdEx = entryPreIndex
					skippy, err := skipPropagationTracker(dAtA[iNdEx:])
					if err != nil {
						return err
					}
					if skippy < 0 {
						return ErrInvalidLengthPropagationTracker
					}
					if (iNdEx + skippy) > postIndex {
						return io.ErrUnexpectedEOF
					}
					iNdEx += skippy
				}
			}
			m.Beacons[mapkey] = *mapvalue
			iNdEx = postIndex
		default:
			iNdEx = preIndex
			skippy, err := skipPropagationTracker(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if skippy < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if (iNdEx + skippy) < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func (m *BeaconDesc) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowPropagationTracker
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: BeaconDesc: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: BeaconDesc: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field PublishedAt", wireType)
			}
			m.PublishedAt = 0
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				m.PublishedAt |= int64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
		case 2:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field DeletedAt", wireType)
			}
			m.DeletedAt = 0
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				m.DeletedAt |= int64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
		case 3:
			if wireType != 2 {
				return fmt.Errorf("proto: wrong wireType = %d for field PublishedBy", wireType)
			}
			var stringLen uint64
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				stringLen |= uint64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			intStringLen := int(stringLen)
			if intStringLen < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			postIndex := iNdEx + intStringLen
			if postIndex < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if postIndex > l {
				return io.ErrUnexpectedEOF
			}
			m.PublishedBy = string(dAtA[iNdEx:postIndex])
			iNdEx = postIndex
		default:
			iNdEx = preIndex
			skippy, err := skipPropagationTracker(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if skippy < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if (iNdEx + skippy) < 0 {
				return ErrInvalidLengthPropagationTracker
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func skipPropagationTracker(dAtA []byte) (n int, err error) {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return 0, ErrIntOverflowPropagationTracker
			}
			if iNdEx >= l {
				return 0, io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= (uint64(b) & 0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		wireType := int(wire & 0x7)
		switch wireType {
		case 0:
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				iNdEx++
				if dAtA[iNdEx-1] < 0x80 {
					break
				}
			}
			return iNdEx, nil
		case 1:
			iNdEx += 8
			return iNdEx, nil
		case 2:
			var length int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowPropagationTracker
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				length |= (int(b) & 0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			if length < 0 {
				return 0, ErrInvalidLengthPropagationTracker
			}
			iNdEx += length
			if iNdEx < 0 {
				return 0, ErrInvalidLengthPropagationTracker
			}
			return iNdEx, nil
		case 3:
			for {
				var innerWire uint64
				var start int = iNdEx
				for shift := uint(0); ; shift += 7 {
					if shift >= 64 {
						return 0, ErrIntOverflowPropagationTracker
					}
					if iNdEx >= l {
						return 0, io.ErrUnexpectedEOF
					}
					b := dAtA[iNdEx]
					iNdEx++
					innerWire |= (uint64(b) & 0x7F) << shift
					if b < 0x80 {
						break
					}
				}
				innerWireType := int(innerWire & 0x7)
				if innerWireType == 4 {
					break
				}
				next, err := skipPropagationTracker(dAtA[start:])
				if err != nil {
					return 0, err
				}
				iNdEx = start + next
				if iNdEx < 0 {
					return 0, ErrInvalidLengthPropagationTracker
				}
			}
			return iNdEx, nil
		case 4:
			return iNdEx, nil
		case 5:
			iNdEx += 4
			return iNdEx, nil
		default:
			return 0, fmt.Errorf("proto: illegal wireType %d", wireType)
		}
	}
	panic("unreachable")
}

var (
	ErrInvalidLengthPropagationTracker = fmt.Errorf("proto: negative length found during unmarshaling")
	ErrIntOverflowPropagationTracker   = fmt.Errorf("proto: integer overflow")
)
