// Code generated by goyacc -l -o promql/parser/generated_parser.y.go promql/parser/generated_parser.y. DO NOT EDIT.
package parser

import __yyfmt__ "fmt"

import (
	"math"
	"strconv"
	"time"

	"github.com/prometheus/prometheus/model/histogram"
	"github.com/prometheus/prometheus/model/labels"
	"github.com/prometheus/prometheus/model/value"
	"github.com/prometheus/prometheus/promql/parser/posrange"

	"github.com/prometheus/common/model"
)

type yySymType struct {
	yys         int
	node        Node
	item        Item
	matchers    []*labels.Matcher
	matcher     *labels.Matcher
	label       labels.Label
	labels      labels.Labels
	lblList     []labels.Label
	strings     []string
	series      []SequenceValue
	histogram   *histogram.FloatHistogram
	descriptors map[string]interface{}
	bucket_set  []float64
	int         int64
	uint        uint64
	float       float64
}

const EQL = 57346
const BLANK = 57347
const COLON = 57348
const COMMA = 57349
const COMMENT = 57350
const DURATION = 57351
const EOF = 57352
const ERROR = 57353
const IDENTIFIER = 57354
const LEFT_BRACE = 57355
const LEFT_BRACKET = 57356
const LEFT_PAREN = 57357
const OPEN_HIST = 57358
const CLOSE_HIST = 57359
const METRIC_IDENTIFIER = 57360
const NUMBER = 57361
const RIGHT_BRACE = 57362
const RIGHT_BRACKET = 57363
const RIGHT_PAREN = 57364
const SEMICOLON = 57365
const SPACE = 57366
const STRING = 57367
const TIMES = 57368
const histogramDescStart = 57369
const SUM_DESC = 57370
const COUNT_DESC = 57371
const SCHEMA_DESC = 57372
const OFFSET_DESC = 57373
const NEGATIVE_OFFSET_DESC = 57374
const BUCKETS_DESC = 57375
const NEGATIVE_BUCKETS_DESC = 57376
const ZERO_BUCKET_DESC = 57377
const ZERO_BUCKET_WIDTH_DESC = 57378
const CUSTOM_VALUES_DESC = 57379
const COUNTER_RESET_HINT_DESC = 57380
const histogramDescEnd = 57381
const operatorsStart = 57382
const ADD = 57383
const DIV = 57384
const EQLC = 57385
const EQL_REGEX = 57386
const GTE = 57387
const GTR = 57388
const TRIM_UPPER = 57389
const TRIM_LOWER = 57390
const LAND = 57391
const LOR = 57392
const LSS = 57393
const LTE = 57394
const LUNLESS = 57395
const MOD = 57396
const MUL = 57397
const NEQ = 57398
const NEQ_REGEX = 57399
const POW = 57400
const SUB = 57401
const AT = 57402
const ATAN2 = 57403
const operatorsEnd = 57404
const aggregatorsStart = 57405
const AVG = 57406
const BOTTOMK = 57407
const COUNT = 57408
const COUNT_VALUES = 57409
const GROUP = 57410
const MAX = 57411
const MIN = 57412
const QUANTILE = 57413
const STDDEV = 57414
const STDVAR = 57415
const SUM = 57416
const TOPK = 57417
const LIMITK = 57418
const LIMIT_RATIO = 57419
const aggregatorsEnd = 57420
const keywordsStart = 57421
const BOOL = 57422
const BY = 57423
const GROUP_LEFT = 57424
const GROUP_RIGHT = 57425
const FILL = 57426
const FILL_LEFT = 57427
const FILL_RIGHT = 57428
const IGNORING = 57429
const OFFSET = 57430
const SMOOTHED = 57431
const ANCHORED = 57432
const ON = 57433
const WITHOUT = 57434
const keywordsEnd = 57435
const preprocessorStart = 57436
const START = 57437
const END = 57438
const STEP = 57439
const RANGE = 57440
const preprocessorEnd = 57441
const counterResetHintsStart = 57442
const UNKNOWN_COUNTER_RESET = 57443
const COUNTER_RESET = 57444
const NOT_COUNTER_RESET = 57445
const GAUGE_TYPE = 57446
const counterResetHintsEnd = 57447
const startSymbolsStart = 57448
const START_METRIC = 57449
const START_SERIES_DESCRIPTION = 57450
const START_EXPRESSION = 57451
const START_METRIC_SELECTOR = 57452
const startSymbolsEnd = 57453

var yyToknames = [...]string{
	"$end",
	"error",
	"$unk",
	"EQL",
	"BLANK",
	"COLON",
	"COMMA",
	"COMMENT",
	"DURATION",
	"EOF",
	"ERROR",
	"IDENTIFIER",
	"LEFT_BRACE",
	"LEFT_BRACKET",
	"LEFT_PAREN",
	"OPEN_HIST",
	"CLOSE_HIST",
	"METRIC_IDENTIFIER",
	"NUMBER",
	"RIGHT_BRACE",
	"RIGHT_BRACKET",
	"RIGHT_PAREN",
	"SEMICOLON",
	"SPACE",
	"STRING",
	"TIMES",
	"histogramDescStart",
	"SUM_DESC",
	"COUNT_DESC",
	"SCHEMA_DESC",
	"OFFSET_DESC",
	"NEGATIVE_OFFSET_DESC",
	"BUCKETS_DESC",
	"NEGATIVE_BUCKETS_DESC",
	"ZERO_BUCKET_DESC",
	"ZERO_BUCKET_WIDTH_DESC",
	"CUSTOM_VALUES_DESC",
	"COUNTER_RESET_HINT_DESC",
	"histogramDescEnd",
	"operatorsStart",
	"ADD",
	"DIV",
	"EQLC",
	"EQL_REGEX",
	"GTE",
	"GTR",
	"TRIM_UPPER",
	"TRIM_LOWER",
	"LAND",
	"LOR",
	"LSS",
	"LTE",
	"LUNLESS",
	"MOD",
	"MUL",
	"NEQ",
	"NEQ_REGEX",
	"POW",
	"SUB",
	"AT",
	"ATAN2",
	"operatorsEnd",
	"aggregatorsStart",
	"AVG",
	"BOTTOMK",
	"COUNT",
	"COUNT_VALUES",
	"GROUP",
	"MAX",
	"MIN",
	"QUANTILE",
	"STDDEV",
	"STDVAR",
	"SUM",
	"TOPK",
	"LIMITK",
	"LIMIT_RATIO",
	"aggregatorsEnd",
	"keywordsStart",
	"BOOL",
	"BY",
	"GROUP_LEFT",
	"GROUP_RIGHT",
	"FILL",
	"FILL_LEFT",
	"FILL_RIGHT",
	"IGNORING",
	"OFFSET",
	"SMOOTHED",
	"ANCHORED",
	"ON",
	"WITHOUT",
	"keywordsEnd",
	"preprocessorStart",
	"START",
	"END",
	"STEP",
	"RANGE",
	"preprocessorEnd",
	"counterResetHintsStart",
	"UNKNOWN_COUNTER_RESET",
	"COUNTER_RESET",
	"NOT_COUNTER_RESET",
	"GAUGE_TYPE",
	"counterResetHintsEnd",
	"startSymbolsStart",
	"START_METRIC",
	"START_SERIES_DESCRIPTION",
	"START_EXPRESSION",
	"START_METRIC_SELECTOR",
	"startSymbolsEnd",
}

var yyStatenames = [...]string{}

const yyEofCode = 1
const yyErrCode = 2
const yyInitialStackSize = 16

var yyExca = [...]int16{
	-1, 1,
	1, -1,
	-2, 0,
	-1, 44,
	1, 166,
	10, 166,
	24, 166,
	-2, 0,
	-1, 76,
	2, 209,
	15, 209,
	81, 209,
	92, 209,
	-2, 120,
	-1, 77,
	2, 210,
	15, 210,
	81, 210,
	92, 210,
	-2, 121,
	-1, 78,
	2, 211,
	15, 211,
	81, 211,
	92, 211,
	-2, 123,
	-1, 79,
	2, 212,
	15, 212,
	81, 212,
	92, 212,
	-2, 124,
	-1, 80,
	2, 213,
	15, 213,
	81, 213,
	92, 213,
	-2, 128,
	-1, 81,
	2, 214,
	15, 214,
	81, 214,
	92, 214,
	-2, 133,
	-1, 82,
	2, 215,
	15, 215,
	81, 215,
	92, 215,
	-2, 135,
	-1, 83,
	2, 216,
	15, 216,
	81, 216,
	92, 216,
	-2, 137,
	-1, 84,
	2, 217,
	15, 217,
	81, 217,
	92, 217,
	-2, 138,
	-1, 85,
	2, 218,
	15, 218,
	81, 218,
	92, 218,
	-2, 139,
	-1, 86,
	2, 219,
	15, 219,
	81, 219,
	92, 219,
	-2, 140,
	-1, 87,
	2, 220,
	15, 220,
	81, 220,
	92, 220,
	-2, 141,
	-1, 88,
	2, 221,
	15, 221,
	81, 221,
	92, 221,
	-2, 145,
	-1, 89,
	2, 222,
	15, 222,
	81, 222,
	92, 222,
	-2, 146,
	-1, 90,
	15, 92,
	-2, 143,
	-1, 91,
	15, 93,
	-2, 144,
	-1, 150,
	41, 293,
	42, 293,
	54, 293,
	55, 293,
	59, 293,
	-2, 22,
	-1, 268,
	9, 278,
	12, 278,
	13, 278,
	18, 278,
	19, 278,
	25, 278,
	41, 278,
	49, 278,
	50, 278,
	53, 278,
	59, 278,
	64, 278,
	65, 278,
	66, 278,
	67, 278,
	68, 278,
	69, 278,
	70, 278,
	71, 278,
	72, 278,
	73, 278,
	74, 278,
	75, 278,
	76, 278,
	77, 278,
	81, 278,
	84, 278,
	85, 278,
	86, 278,
	88, 278,
	89, 278,
	90, 278,
	92, 278,
	95, 278,
	96, 278,
	97, 278,
	98, 278,
	-2, 0,
	-1, 269,
	9, 278,
	12, 278,
	13, 278,
	18, 278,
	19, 278,
	25, 278,
	41, 278,
	49, 278,
	50, 278,
	53, 278,
	59, 278,
	64, 278,
	65, 278,
	66, 278,
	67, 278,
	68, 278,
	69, 278,
	70, 278,
	71, 278,
	72, 278,
	73, 278,
	74, 278,
	75, 278,
	76, 278,
	77, 278,
	81, 278,
	84, 278,
	85, 278,
	86, 278,
	88, 278,
	89, 278,
	90, 278,
	92, 278,
	95, 278,
	96, 278,
	97, 278,
	98, 278,
	-2, 0,
}

const yyPrivate = 57344

const yyLast = 1411

var yyAct = [...]int16{
	61, 377, 200, 445, 203, 447, 365, 307, 257, 211,
	452, 51, 101, 155, 74, 383, 99, 241, 428, 64,
	429, 396, 384, 134, 138, 71, 68, 136, 68, 204,
	171, 135, 199, 137, 199, 139, 67, 405, 67, 442,
	395, 459, 460, 461, 462, 73, 140, 266, 441, 68,
	359, 267, 128, 268, 269, 291, 407, 355, 92, 67,
	92, 263, 264, 265, 474, 360, 6, 356, 165, 208,
	148, 405, 150, 463, 68, 357, 93, 446, 93, 207,
	354, 92, 413, 361, 67, 431, 198, 197, 198, 197,
	209, 141, 142, 143, 145, 232, 122, 412, 127, 93,
	46, 209, 436, 430, 411, 348, 92, 157, 210, 198,
	197, 398, 138, 397, 216, 217, 216, 217, 151, 210,
	347, 435, 414, 139, 93, 196, 124, 126, 125, 213,
	218, 219, 220, 221, 222, 223, 191, 288, 289, 195,
	214, 214, 214, 214, 214, 214, 214, 201, 393, 391,
	192, 390, 261, 237, 215, 215, 215, 215, 215, 215,
	215, 226, 229, 224, 214, 147, 353, 149, 225, 212,
	259, 2, 3, 4, 5, 247, 68, 249, 215, 453,
	253, 68, 199, 228, 378, 132, 67, 94, 388, 352,
	387, 67, 190, 292, 251, 260, 227, 133, 131, 68,
	296, 134, 480, 252, 287, 199, 202, 290, 92, 67,
	476, 214, 386, 208, 298, 299, 292, 250, 215, 42,
	166, 129, 132, 207, 304, 215, 93, 214, 401, 129,
	132, 92, 479, 400, 133, 131, 198, 197, 134, 130,
	208, 215, 133, 131, 303, 209, 134, 130, 399, 93,
	207, 478, 157, 351, 477, 242, 297, 243, 295, 198,
	197, 294, 262, 210, 216, 217, 379, 380, 270, 271,
	272, 273, 274, 275, 276, 277, 278, 279, 280, 281,
	282, 283, 284, 285, 286, 385, 358, 193, 194, 381,
	382, 293, 389, 140, 152, 146, 392, 245, 144, 205,
	206, 7, 213, 214, 10, 243, 425, 214, 424, 244,
	246, 394, 255, 214, 96, 423, 208, 215, 422, 258,
	467, 215, 421, 261, 420, 419, 207, 215, 172, 173,
	174, 175, 176, 177, 178, 179, 180, 181, 182, 183,
	184, 185, 186, 187, 188, 245, 434, 418, 209, 417,
	403, 416, 415, 164, 129, 132, 8, 244, 246, 439,
	44, 214, 404, 406, 158, 408, 210, 133, 131, 409,
	410, 134, 130, 208, 156, 215, 105, 43, 129, 132,
	129, 132, 350, 207, 302, 481, 1, 103, 301, 432,
	426, 133, 131, 133, 131, 134, 130, 134, 130, 189,
	214, 59, 75, 300, 427, 209, 58, 95, 438, 57,
	163, 240, 162, 440, 215, 102, 239, 433, 444, 448,
	449, 450, 161, 210, 451, 100, 455, 160, 457, 238,
	465, 466, 454, 248, 456, 468, 129, 132, 103, 464,
	159, 310, 56, 55, 402, 170, 214, 469, 470, 133,
	131, 323, 471, 134, 130, 54, 473, 329, 236, 72,
	215, 437, 9, 9, 309, 53, 52, 50, 475, 129,
	132, 168, 230, 49, 169, 48, 167, 129, 132, 214,
	482, 47, 133, 131, 60, 256, 134, 130, 325, 326,
	133, 131, 327, 215, 134, 130, 472, 362, 349, 104,
	340, 254, 305, 311, 313, 315, 316, 317, 328, 330,
	333, 334, 335, 336, 337, 341, 342, 97, 458, 312,
	314, 318, 319, 320, 321, 322, 324, 331, 346, 345,
	332, 310, 129, 132, 338, 339, 343, 344, 308, 62,
	153, 323, 0, 0, 0, 133, 131, 329, 364, 134,
	130, 306, 0, 0, 309, 363, 0, 0, 0, 367,
	368, 366, 373, 375, 372, 374, 369, 370, 371, 376,
	0, 0, 0, 0, 0, 0, 0, 0, 325, 326,
	0, 0, 327, 0, 0, 0, 0, 0, 0, 0,
	340, 0, 0, 311, 313, 315, 316, 317, 328, 330,
	333, 334, 335, 336, 337, 341, 342, 0, 0, 312,
	314, 318, 319, 320, 321, 322, 324, 331, 346, 345,
	332, 0, 0, 0, 338, 339, 343, 344, 68, 0,
	0, 63, 94, 0, 69, 443, 0, 25, 67, 0,
	0, 234, 0, 0, 70, 0, 367, 368, 366, 373,
	375, 372, 374, 369, 370, 371, 376, 158, 0, 0,
	92, 0, 0, 0, 0, 0, 0, 156, 21, 22,
	0, 0, 23, 0, 0, 154, 0, 0, 93, 0,
	103, 0, 0, 76, 77, 78, 79, 80, 81, 82,
	83, 84, 85, 86, 87, 88, 89, 0, 0, 0,
	13, 0, 102, 16, 17, 18, 0, 27, 41, 40,
	0, 33, 100, 0, 90, 91, 65, 66, 68, 0,
	98, 63, 94, 0, 69, 103, 0, 25, 67, 0,
	0, 233, 0, 0, 70, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	92, 0, 0, 0, 0, 0, 0, 0, 21, 22,
	0, 0, 23, 0, 0, 0, 0, 0, 93, 0,
	0, 0, 0, 76, 77, 78, 79, 80, 81, 82,
	83, 84, 85, 86, 87, 88, 89, 0, 0, 0,
	13, 0, 0, 16, 17, 18, 0, 27, 41, 40,
	0, 33, 0, 0, 90, 91, 65, 66, 68, 0,
	0, 63, 94, 0, 69, 0, 0, 25, 67, 0,
	0, 231, 0, 0, 70, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	92, 0, 0, 0, 0, 0, 0, 0, 21, 22,
	0, 0, 23, 0, 0, 0, 0, 0, 93, 0,
	0, 0, 0, 76, 77, 78, 79, 80, 81, 82,
	83, 84, 85, 86, 87, 88, 89, 0, 0, 0,
	13, 0, 0, 16, 17, 18, 0, 27, 41, 40,
	0, 33, 0, 0, 90, 91, 65, 66, 68, 45,
	0, 63, 94, 0, 69, 0, 0, 25, 67, 0,
	0, 0, 0, 0, 70, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	92, 0, 0, 0, 0, 0, 0, 0, 21, 22,
	0, 0, 23, 0, 0, 0, 0, 0, 93, 0,
	0, 0, 0, 76, 77, 78, 79, 80, 81, 82,
	83, 84, 85, 86, 87, 88, 89, 0, 0, 0,
	13, 0, 0, 16, 17, 18, 0, 27, 41, 40,
	0, 33, 0, 0, 90, 91, 65, 66, 68, 0,
	0, 63, 94, 0, 69, 0, 0, 25, 67, 0,
	0, 0, 0, 0, 70, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	92, 0, 0, 0, 0, 0, 0, 0, 21, 22,
	0, 0, 23, 0, 0, 0, 0, 0, 93, 0,
	0, 0, 128, 76, 77, 78, 79, 80, 81, 82,
	83, 84, 85, 86, 87, 88, 89, 0, 0, 0,
	13, 0, 0, 16, 17, 18, 0, 27, 41, 40,
	108, 33, 20, 94, 90, 91, 65, 66, 25, 0,
	0, 0, 119, 120, 0, 0, 122, 0, 127, 107,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 21,
	22, 0, 0, 23, 0, 0, 124, 126, 125, 0,
	0, 0, 0, 0, 11, 12, 14, 15, 19, 24,
	26, 28, 29, 30, 31, 32, 36, 37, 0, 0,
	0, 13, 0, 0, 16, 17, 18, 0, 27, 41,
	40, 0, 33, 20, 42, 34, 35, 38, 39, 25,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	21, 22, 0, 0, 23, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 11, 12, 14, 15, 19,
	24, 26, 28, 29, 30, 31, 32, 36, 37, 0,
	128, 0, 13, 0, 0, 16, 17, 18, 235, 27,
	41, 40, 0, 33, 0, 0, 34, 35, 38, 39,
	0, 128, 0, 0, 0, 0, 0, 106, 108, 109,
	0, 110, 111, 112, 113, 114, 115, 116, 117, 118,
	119, 120, 121, 0, 122, 123, 127, 107, 106, 108,
	109, 0, 110, 111, 112, 113, 114, 115, 116, 117,
	118, 119, 120, 121, 0, 122, 123, 127, 107, 0,
	0, 128, 0, 0, 124, 126, 125, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 128, 0, 0, 124, 126, 125, 106, 108,
	109, 0, 110, 111, 112, 113, 114, 0, 116, 117,
	118, 119, 120, 121, 128, 122, 123, 127, 107, 106,
	108, 109, 0, 110, 111, 112, 113, 0, 0, 116,
	117, 0, 119, 120, 121, 0, 122, 123, 127, 107,
	0, 106, 108, 0, 0, 124, 126, 125, 0, 0,
	0, 0, 0, 0, 119, 120, 0, 0, 122, 123,
	127, 107, 0, 0, 0, 0, 124, 126, 125, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 124, 126,
	125,
}

var yyPact = [...]int16{
	64, 291, 1141, 1141, 889, 1060, -1000, -1000, -1000, 206,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, 700, -1000, 374, -1000, 1227, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, 428, 31, 278, 278, 283, 280, -1000, -1000, 979,
	-1000, 979, 174, -1000, 279, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, 655, -1000, -1000, 420, -1000, -1000,
	408, 349, -1000, -1000, 44, -1000, -50, -50, -50, -50,
	-50, -50, -50, -50, -50, -50, -50, -50, -50, -50,
	-50, -50, -50, -50, 190, -1000, -1000, 204, 167, 17,
	17, 17, 17, 17, 17, 278, -57, -1000, 181, 181,
	799, -1000, -1000, -1000, 709, -1000, 619, 1206, 436, 38,
	-35, -1000, 17, 409, -1000, -1000, 253, 301, -1000, -1000,
	413, -1000, 192, -1000, 178, 307, 979, -1000, -23, -40,
	-29, -1000, 979, 979, 979, 979, 979, 979, 979, 979,
	979, 979, 979, 979, 979, 979, 979, 979, 979, -1000,
	-1000, -1000, 40, 276, 246, 243, 428, -1000, -1000, 17,
	-1000, 241, -1000, -1000, -1000, -1000, -1000, -1000, -1000, 231,
	231, 382, -1000, 428, -1000, 17, 229, 209, 143, 143,
	-35, -35, -35, -35, -1000, -1000, -1000, 529, -1000, -1000,
	98, -1000, 1227, -1000, -1000, -1000, -1000, 491, -1000, 362,
	-1000, 164, -1000, -1000, -1000, -1000, -1000, 55, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, 41, 49, 24, -1000, -1000,
	-1000, 531, 1028, 169, 169, 169, 181, 181, 181, 181,
	38, 38, 1320, 1320, 1320, 1320, 1320, 1298, 1277, 1320,
	1320, 1298, 38, 38, 1320, 38, 1028, -1000, 197, 175,
	173, 17, -35, 129, 127, 17, 436, 126, -1000, -1000,
	-1000, 19, -1000, 91, 89, 226, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, -1000, -1000, -1000, -1000, 979, 17,
	-1000, -1000, -1000, -1000, -1000, -1000, 52, 52, 30, 52,
	136, 136, 80, 105, -1000, -1000, 346, 345, 343, 341,
	319, 318, 316, 312, 309, 302, 300, -1000, 65, -68,
	-65, -1000, -1000, -1000, -1000, -1000, 81, 63, 17, 395,
	-1000, -1000, 339, -1000, 100, -1000, -1000, -1000, -1000, -1000,
	439, -1000, 1227, 337, -1000, -1000, -1000, 52, -1000, 22,
	13, 618, -1000, -1000, -1000, 18, 60, 60, 60, 231,
	165, 165, 18, 165, 18, -60, 51, 172, 169, 169,
	-1000, -1000, 313, -1000, 17, -1000, -1000, -1000, -1000, -1000,
	-1000, 52, 52, -1000, -1000, -1000, 52, -1000, -1000, -1000,
	-1000, -1000, -1000, 60, -1000, -1000, -1000, -1000, -1000, -1000,
	-1000, -1000, -1000, -1000, 42, -1000, -1000, 17, 188, -1000,
	-1000, -1000, 230, -1000, -1000, 180, -1000, 364, -1000, -1000,
	-1000, -1000, -1000,
}

var yyPgo = [...]int16{
	0, 540, 13, 539, 7, 17, 538, 459, 25, 19,
	12, 518, 14, 304, 356, 517, 16, 502, 22, 15,
	501, 499, 8, 497, 6, 10, 496, 3, 5, 4,
	485, 29, 2, 484, 481, 31, 220, 476, 475, 474,
	95, 473, 472, 27, 471, 1, 45, 467, 11, 466,
	465, 455, 445, 443, 442, 409, 406, 401, 0, 402,
	9, 399, 386, 377,
}

var yyR1 = [...]int8{
	0, 62, 62, 62, 62, 62, 62, 62, 40, 40,
	40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
	40, 40, 40, 34, 34, 34, 34, 35, 35, 38,
	38, 38, 38, 38, 38, 38, 38, 38, 38, 38,
	38, 38, 38, 38, 38, 38, 38, 36, 39, 39,
	52, 52, 44, 44, 44, 44, 37, 37, 37, 37,
	37, 37, 18, 18, 18, 18, 17, 17, 17, 4,
	4, 4, 45, 45, 41, 41, 41, 41, 43, 43,
	42, 42, 42, 53, 60, 49, 49, 50, 51, 33,
	33, 33, 9, 9, 47, 55, 55, 55, 55, 55,
	55, 56, 57, 57, 57, 46, 46, 46, 1, 1,
	1, 2, 2, 2, 2, 2, 2, 2, 14, 14,
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
	7, 13, 13, 13, 13, 15, 15, 15, 16, 16,
	16, 16, 16, 16, 16, 63, 21, 21, 21, 21,
	20, 20, 20, 20, 20, 20, 20, 20, 20, 30,
	30, 30, 22, 22, 22, 22, 23, 23, 23, 24,
	24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
	25, 25, 26, 26, 26, 11, 11, 11, 11, 3,
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	3, 3, 3, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6, 6, 8,
	8, 5, 5, 5, 5, 48, 48, 29, 29, 31,
	31, 32, 32, 28, 27, 27, 54, 10, 19, 19,
	61, 61, 61, 61, 61, 61, 61, 61, 61, 61,
	12, 12, 58, 58, 58, 58, 58, 58, 58, 58,
	58, 58, 58, 58, 59,
}

var yyR2 = [...]int8{
	0, 2, 2, 2, 2, 2, 2, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 3, 3, 2, 2, 2, 2, 4,
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
	4, 4, 4, 4, 4, 4, 4, 1, 0, 1,
	3, 3, 1, 1, 3, 3, 1, 3, 3, 3,
	5, 5, 3, 4, 2, 1, 3, 1, 2, 1,
	1, 1, 3, 4, 2, 2, 2, 2, 3, 2,
	3, 1, 2, 3, 1, 3, 3, 2, 2, 3,
	5, 3, 1, 1, 4, 6, 5, 6, 5, 4,
	3, 2, 2, 1, 1, 3, 4, 2, 3, 1,
	2, 3, 3, 1, 3, 3, 2, 1, 2, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 3, 4, 2, 0, 3, 1, 2, 3, 3,
	1, 3, 3, 2, 1, 2, 0, 3, 2, 1,
	1, 3, 1, 3, 4, 1, 3, 5, 5, 1,
	1, 1, 4, 3, 3, 2, 3, 1, 2, 3,
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	4, 3, 3, 1, 2, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 2,
	2, 1, 1, 1, 2, 1, 1, 1, 0, 1,
	1, 2, 3, 3, 4, 4, 6, 7, 4, 1,
	1, 1, 1, 2, 3, 3, 3, 3, 3, 3,
	3, 3, 6, 1, 3,
}

var yyChk = [...]int16{
	-1000, -62, 107, 108, 109, 110, 2, 10, -14, -7,
	-13, 64, 65, 81, 66, 67, 84, 85, 86, 68,
	12, 49, 50, 53, 69, 18, 70, 88, 71, 72,
	73, 74, 75, 92, 95, 96, 76, 77, 97, 98,
	90, 89, 13, -63, -14, 10, -40, -34, -38, -41,
	-47, -48, -49, -50, -51, -53, -54, -55, -56, -57,
	-33, -58, -3, 12, -9, 97, 98, 19, 9, 15,
	25, -8, -7, -46, -12, -59, 64, 65, 66, 67,
	68, 69, 70, 71, 72, 73, 74, 75, 76, 77,
	95, 96, 41, 59, 13, -57, -13, -15, 20, -16,
	12, -10, 2, 25, -21, 2, 41, 61, 42, 43,
	45, 46, 47, 48, 49, 50, 51, 52, 53, 54,
	55, 56, 58, 59, 88, 90, 89, 60, 14, 41,
	59, 55, 42, 54, 58, -35, -43, 2, 81, 92,
	15, -43, -43, -43, 15, -43, 15, -40, -58, -40,
	-58, -46, 15, -1, 20, -2, 12, -10, 2, 20,
	7, 2, 4, 2, 4, 24, -36, -37, -44, -39,
	-52, 80, -36, -36, -36, -36, -36, -36, -36, -36,
	-36, -36, -36, -36, -36, -36, -36, -36, -36, -61,
	2, -48, -8, 97, 98, -12, -58, 70, 69, 15,
	-32, -9, 2, -29, -31, 95, 96, 19, 9, 41,
	59, -60, 2, -58, -48, -8, 97, 98, -58, -58,
	-58, -58, -58, -58, -43, -35, -18, 15, 2, -18,
	-42, 22, -40, 22, 22, 22, 22, -58, 20, 7,
	2, -5, 2, 4, 56, 44, 57, -5, 20, -16,
	25, 2, 25, 2, -20, 5, -30, -22, 12, -29,
	-31, 16, -40, 84, 85, 86, 87, 91, 82, 83,
	-40, -40, -40, -40, -40, -40, -40, -40, -40, -40,
	-40, -40, -40, -40, -40, -40, -40, -48, 97, 98,
	-12, 15, -58, 15, 15, 15, -58, 15, -29, -29,
	21, 6, 2, 15, 15, -17, 22, -4, -6, 25,
	2, 64, 80, 65, 81, 66, 67, 68, 82, 83,
	84, 85, 86, 12, 87, 49, 50, 53, 69, 18,
	70, 88, 91, 71, 72, 73, 74, 75, 95, 96,
	61, 76, 77, 97, 98, 90, 89, 22, 7, 7,
	20, -2, 25, 2, 25, 2, 26, 26, -31, 26,
	41, 59, -23, 24, 17, -24, 30, 28, 29, 35,
	36, 37, 33, 31, 34, 32, 38, -45, 15, -45,
	-45, -18, -18, -19, -18, -19, 15, 15, 15, -58,
	22, 22, -58, 22, -60, 21, 2, 22, 22, 22,
	7, 2, -40, -58, -28, 19, -28, 26, -28, -22,
	-22, 24, 17, 2, 17, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, -48, -8, 86, 85,
	22, 22, -58, 22, 7, 21, 2, 22, -4, 22,
	-28, 26, 26, 17, -24, -27, 59, -28, -32, -32,
	-32, -29, -25, 14, -25, -27, -25, -27, -11, 101,
	102, 103, 104, 22, -48, -45, -45, 7, -58, -28,
	-28, -28, -26, -32, 22, -58, 22, 24, 21, 2,
	22, 21, -32,
}

var yyDef = [...]int16{
	0, -2, 154, 154, 0, 0, 7, 6, 1, 154,
	119, 120, 121, 122, 123, 124, 125, 126, 127, 128,
	129, 130, 131, 132, 133, 134, 135, 136, 137, 138,
	139, 140, 141, 142, 143, 144, 145, 146, 147, 148,
	149, 150, 0, 2, -2, 3, 4, 8, 9, 10,
	11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
	21, 22, 0, 129, 0, 147, 148, 265, 266, 0,
	276, 0, 103, 104, 0, 303, -2, -2, -2, -2,
	-2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
	-2, -2, 259, 260, 0, 5, 118, 0, 153, 156,
	0, 160, 164, 277, 165, 169, 48, 48, 48, 48,
	48, 48, 48, 48, 48, 48, 48, 48, 48, 48,
	48, 48, 48, 48, 0, 87, 88, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 25, 26, 0, 0,
	0, 74, 75, 76, 0, 77, 0, 0, 22, 101,
	-2, 102, 0, 0, 107, 109, 0, 113, 117, 151,
	0, 157, 0, 163, 0, 168, 0, 47, 56, 52,
	53, 49, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 85,
	86, 280, 0, 0, 0, 0, 289, 290, 291, 0,
	89, 0, 91, 271, 272, 92, 93, 267, 268, 0,
	0, 0, 100, 84, 292, 0, 0, 0, 294, 295,
	296, 297, 298, 299, 23, 24, 27, 0, 65, 28,
	0, 79, 81, 79, 79, 83, 304, 0, 105, 0,
	110, 0, 116, 261, 262, 263, 264, 0, 152, 155,
	158, 161, 159, 162, 167, 170, 172, 175, 179, 180,
	181, 0, 29, 0, 0, 0, 0, 0, -2, -2,
	30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
	40, 41, 42, 43, 44, 45, 46, 281, 0, 0,
	0, 0, 293, 0, 0, 0, 0, 0, 269, 270,
	94, 0, 99, 0, 0, 0, 64, 67, 69, 70,
	71, 223, 224, 225, 226, 227, 228, 229, 230, 231,
	232, 233, 234, 235, 236, 237, 238, 239, 240, 241,
	242, 243, 244, 245, 246, 247, 248, 249, 250, 251,
	252, 253, 254, 255, 256, 257, 258, 78, 82, 0,
	106, 108, 111, 115, 112, 114, 0, 0, 0, 0,
	0, 0, 0, 0, 185, 187, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 57, 0, 58,
	59, 50, 51, 54, 279, 55, 0, 0, 0, 0,
	282, 283, 0, 90, 0, 96, 98, 300, 301, 62,
	0, 68, 80, 0, 171, 273, 173, 0, 176, 0,
	0, 0, 183, 188, 184, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	284, 285, 0, 288, 0, 95, 97, 63, 66, 302,
	174, 0, 0, 182, 186, 189, 0, 275, 190, 191,
	192, 193, 194, 0, 195, 196, 197, 198, 199, 205,
	206, 207, 208, 72, 0, 60, 61, 0, 0, 177,
	178, 274, 0, 203, 73, 0, 286, 0, 201, 204,
	287, 200, 202,
}

var yyTok1 = [...]int8{
	1,
}

var yyTok2 = [...]int8{
	2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
	12, 13, 14, 15, 16, 17, 18, 19, 20, 21,
	22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
	32, 33, 34, 35, 36, 37, 38, 39, 40, 41,
	42, 43, 44, 45, 46, 47, 48, 49, 50, 51,
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61,
	62, 63, 64, 65, 66, 67, 68, 69, 70, 71,
	72, 73, 74, 75, 76, 77, 78, 79, 80, 81,
	82, 83, 84, 85, 86, 87, 88, 89, 90, 91,
	92, 93, 94, 95, 96, 97, 98, 99, 100, 101,
	102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
}

var yyTok3 = [...]int8{
	0,
}

var yyErrorMessages = [...]struct {
	state int
	token int
	msg   string
}{}

/*	parser for yacc output	*/

var (
	yyDebug        = 0
	yyErrorVerbose = false
)

type yyLexer interface {
	Lex(lval *yySymType) int
	Error(s string)
}

type yyParser interface {
	Parse(yyLexer) int
	Lookahead() int
}

type yyParserImpl struct {
	lval  yySymType
	stack [yyInitialStackSize]yySymType
	char  int
}

func (p *yyParserImpl) Lookahead() int {
	return p.char
}

func yyNewParser() yyParser {
	return &yyParserImpl{}
}

const yyFlag = -1000

func yyTokname(c int) string {
	if c >= 1 && c-1 < len(yyToknames) {
		if yyToknames[c-1] != "" {
			return yyToknames[c-1]
		}
	}
	return __yyfmt__.Sprintf("tok-%v", c)
}

func yyStatname(s int) string {
	if s >= 0 && s < len(yyStatenames) {
		if yyStatenames[s] != "" {
			return yyStatenames[s]
		}
	}
	return __yyfmt__.Sprintf("state-%v", s)
}

func yyErrorMessage(state, lookAhead int) string {
	const TOKSTART = 4

	if !yyErrorVerbose {
		return "syntax error"
	}

	for _, e := range yyErrorMessages {
		if e.state == state && e.token == lookAhead {
			return "syntax error: " + e.msg
		}
	}

	res := "syntax error: unexpected " + yyTokname(lookAhead)

	// To match Bison, suggest at most four expected tokens.
	expected := make([]int, 0, 4)

	// Look for shiftable tokens.
	base := int(yyPact[state])
	for tok := TOKSTART; tok-1 < len(yyToknames); tok++ {
		if n := base + tok; n >= 0 && n < yyLast && int(yyChk[int(yyAct[n])]) == tok {
			if len(expected) == cap(expected) {
				return res
			}
			expected = append(expected, tok)
		}
	}

	if yyDef[state] == -2 {
		i := 0
		for yyExca[i] != -1 || int(yyExca[i+1]) != state {
			i += 2
		}

		// Look for tokens that we accept or reduce.
		for i += 2; yyExca[i] >= 0; i += 2 {
			tok := int(yyExca[i])
			if tok < TOKSTART || yyExca[i+1] == 0 {
				continue
			}
			if len(expected) == cap(expected) {
				return res
			}
			expected = append(expected, tok)
		}

		// If the default action is to accept or reduce, give up.
		if yyExca[i+1] != 0 {
			return res
		}
	}

	for i, tok := range expected {
		if i == 0 {
			res += ", expecting "
		} else {
			res += " or "
		}
		res += yyTokname(tok)
	}
	return res
}

func yylex1(lex yyLexer, lval *yySymType) (char, token int) {
	token = 0
	char = lex.Lex(lval)
	if char <= 0 {
		token = int(yyTok1[0])
		goto out
	}
	if char < len(yyTok1) {
		token = int(yyTok1[char])
		goto out
	}
	if char >= yyPrivate {
		if char < yyPrivate+len(yyTok2) {
			token = int(yyTok2[char-yyPrivate])
			goto out
		}
	}
	for i := 0; i < len(yyTok3); i += 2 {
		token = int(yyTok3[i+0])
		if token == char {
			token = int(yyTok3[i+1])
			goto out
		}
	}

out:
	if token == 0 {
		token = int(yyTok2[1]) /* unknown char */
	}
	if yyDebug >= 3 {
		__yyfmt__.Printf("lex %s(%d)\n", yyTokname(token), uint(char))
	}
	return char, token
}

func yyParse(yylex yyLexer) int {
	return yyNewParser().Parse(yylex)
}

func (yyrcvr *yyParserImpl) Parse(yylex yyLexer) int {
	var yyn int
	var yyVAL yySymType
	var yyDollar []yySymType
	_ = yyDollar // silence set and not used
	yyS := yyrcvr.stack[:]

	Nerrs := 0   /* number of errors */
	Errflag := 0 /* error recovery flag */
	yystate := 0
	yyrcvr.char = -1
	yytoken := -1 // yyrcvr.char translated into internal numbering
	defer func() {
		// Make sure we report no lookahead when not parsing.
		yystate = -1
		yyrcvr.char = -1
		yytoken = -1
	}()
	yyp := -1
	goto yystack

ret0:
	return 0

ret1:
	return 1

yystack:
	/* put a state and value onto the stack */
	if yyDebug >= 4 {
		__yyfmt__.Printf("char %v in %v\n", yyTokname(yytoken), yyStatname(yystate))
	}

	yyp++
	if yyp >= len(yyS) {
		nyys := make([]yySymType, len(yyS)*2)
		copy(nyys, yyS)
		yyS = nyys
	}
	yyS[yyp] = yyVAL
	yyS[yyp].yys = yystate

yynewstate:
	yyn = int(yyPact[yystate])
	if yyn <= yyFlag {
		goto yydefault /* simple state */
	}
	if yyrcvr.char < 0 {
		yyrcvr.char, yytoken = yylex1(yylex, &yyrcvr.lval)
	}
	yyn += yytoken
	if yyn < 0 || yyn >= yyLast {
		goto yydefault
	}
	yyn = int(yyAct[yyn])
	if int(yyChk[yyn]) == yytoken { /* valid shift */
		yyrcvr.char = -1
		yytoken = -1
		yyVAL = yyrcvr.lval
		yystate = yyn
		if Errflag > 0 {
			Errflag--
		}
		goto yystack
	}

yydefault:
	/* default state action */
	yyn = int(yyDef[yystate])
	if yyn == -2 {
		if yyrcvr.char < 0 {
			yyrcvr.char, yytoken = yylex1(yylex, &yyrcvr.lval)
		}

		/* look through exception table */
		xi := 0
		for {
			if yyExca[xi+0] == -1 && int(yyExca[xi+1]) == yystate {
				break
			}
			xi += 2
		}
		for xi += 2; ; xi += 2 {
			yyn = int(yyExca[xi+0])
			if yyn < 0 || yyn == yytoken {
				break
			}
		}
		yyn = int(yyExca[xi+1])
		if yyn < 0 {
			goto ret0
		}
	}
	if yyn == 0 {
		/* error ... attempt to resume parsing */
		switch Errflag {
		case 0: /* brand new error */
			yylex.Error(yyErrorMessage(yystate, yytoken))
			Nerrs++
			if yyDebug >= 1 {
				__yyfmt__.Printf("%s", yyStatname(yystate))
				__yyfmt__.Printf(" saw %s\n", yyTokname(yytoken))
			}
			fallthrough

		case 1, 2: /* incompletely recovered error ... try again */
			Errflag = 3

			/* find a state where "error" is a legal shift action */
			for yyp >= 0 {
				yyn = int(yyPact[yyS[yyp].yys]) + yyErrCode
				if yyn >= 0 && yyn < yyLast {
					yystate = int(yyAct[yyn]) /* simulate a shift of "error" */
					if int(yyChk[yystate]) == yyErrCode {
						goto yystack
					}
				}

				/* the current p has no shift on "error", pop stack */
				if yyDebug >= 2 {
					__yyfmt__.Printf("error recovery pops state %d\n", yyS[yyp].yys)
				}
				yyp--
			}
			/* there is no state on the stack with an error shift ... abort */
			goto ret1

		case 3: /* no shift yet; clobber input char */
			if yyDebug >= 2 {
				__yyfmt__.Printf("error recovery discards %s\n", yyTokname(yytoken))
			}
			if yytoken == yyEofCode {
				goto ret1
			}
			yyrcvr.char = -1
			yytoken = -1
			goto yynewstate /* try again in the same state */
		}
	}

	/* reduction by production yyn */
	if yyDebug >= 2 {
		__yyfmt__.Printf("reduce %v in:\n\t%v\n", yyn, yyStatname(yystate))
	}

	yynt := yyn
	yypt := yyp
	_ = yypt // guard against "declared and not used"

	yyp -= int(yyR2[yyn])
	// yyp is now the index of $0. Perform the default action. Iff the
	// reduced production is ε, $1 is possibly out of range.
	if yyp+1 >= len(yyS) {
		nyys := make([]yySymType, len(yyS)*2)
		copy(nyys, yyS)
		yyS = nyys
	}
	yyVAL = yyS[yyp+1]

	/* consult goto table to find next state */
	yyn = int(yyR1[yyn])
	yyg := int(yyPgo[yyn])
	yyj := yyg + yyS[yyp].yys + 1

	if yyj >= yyLast {
		yystate = int(yyAct[yyg])
	} else {
		yystate = int(yyAct[yyj])
		if int(yyChk[yystate]) != -yyn {
			yystate = int(yyAct[yyg])
		}
	}
	// dummy call; replaced with literal code
	switch yynt {

	case 1:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).generatedParserResult = yyDollar[2].labels
		}
	case 3:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).addParseErrf(posrange.PositionRange{}, "no expression found in input")
		}
	case 4:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).generatedParserResult = yyDollar[2].node
		}
	case 5:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).generatedParserResult = yyDollar[2].node
		}
	case 7:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("", "")
		}
	case 23:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newAggregateExpr(yyDollar[1].item, yyDollar[2].node, yyDollar[3].node, false)
		}
	case 24:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newAggregateExpr(yyDollar[1].item, yyDollar[3].node, yyDollar[2].node, false)
		}
	case 25:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newAggregateExpr(yyDollar[1].item, &AggregateExpr{}, yyDollar[2].node, true)
		}
	case 26:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("aggregation", "")
			yyVAL.node = yylex.(*parser).newAggregateExpr(yyDollar[1].item, &AggregateExpr{}, Expressions{}, false)
		}
	case 27:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.node = &AggregateExpr{
				Grouping: yyDollar[2].strings,
			}
		}
	case 28:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.node = &AggregateExpr{
				Grouping: yyDollar[2].strings,
				Without:  true,
			}
		}
	case 29:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 30:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 31:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 32:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 33:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 34:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 35:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 36:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 37:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 38:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 39:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 40:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 41:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 42:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 43:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 44:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 45:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 46:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = yylex.(*parser).newBinaryExpression(yyDollar[1].node, yyDollar[2].item, yyDollar[3].node, yyDollar[4].node)
		}
	case 48:
		yyDollar = yyS[yypt-0 : yypt+1]
		{
			yyVAL.node = &BinaryExpr{
				VectorMatching: &VectorMatching{Card: CardOneToOne},
			}
		}
	case 49:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.node = &BinaryExpr{
				VectorMatching: &VectorMatching{Card: CardOneToOne},
				ReturnBool:     true,
			}
		}
	case 50:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			yyVAL.node.(*BinaryExpr).VectorMatching.MatchingLabels = yyDollar[3].strings
		}
	case 51:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			yyVAL.node.(*BinaryExpr).VectorMatching.MatchingLabels = yyDollar[3].strings
			yyVAL.node.(*BinaryExpr).VectorMatching.On = true
		}
	case 54:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			yyVAL.node.(*BinaryExpr).VectorMatching.Card = CardManyToOne
			yyVAL.node.(*BinaryExpr).VectorMatching.Include = yyDollar[3].strings
		}
	case 55:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			yyVAL.node.(*BinaryExpr).VectorMatching.Card = CardOneToMany
			yyVAL.node.(*BinaryExpr).VectorMatching.Include = yyDollar[3].strings
		}
	case 57:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			fill := yyDollar[3].node.(*NumberLiteral).Val
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.LHS = &fill
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.RHS = &fill
		}
	case 58:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			fill := yyDollar[3].node.(*NumberLiteral).Val
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.LHS = &fill
		}
	case 59:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			fill := yyDollar[3].node.(*NumberLiteral).Val
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.RHS = &fill
		}
	case 60:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			yyVAL.node = yyDollar[1].node
			fill_left := yyDollar[3].node.(*NumberLiteral).Val
			fill_right := yyDollar[5].node.(*NumberLiteral).Val
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.LHS = &fill_left
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.RHS = &fill_right
		}
	case 61:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			fill_right := yyDollar[3].node.(*NumberLiteral).Val
			fill_left := yyDollar[5].node.(*NumberLiteral).Val
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.LHS = &fill_left
			yyVAL.node.(*BinaryExpr).VectorMatching.FillValues.RHS = &fill_right
		}
	case 62:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.strings = yyDollar[2].strings
		}
	case 63:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.strings = yyDollar[2].strings
		}
	case 64:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.strings = []string{}
		}
	case 65:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("grouping opts", "\"(\"")
			yyVAL.strings = nil
		}
	case 66:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.strings = append(yyDollar[1].strings, yyDollar[3].item.Val)
		}
	case 67:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.strings = []string{yyDollar[1].item.Val}
		}
	case 68:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("grouping opts", "\",\" or \")\"")
			yyVAL.strings = yyDollar[1].strings
		}
	case 69:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			if !model.UTF8Validation.IsValidLabelName(yyDollar[1].item.Val) {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "invalid label name for grouping: %q", yyDollar[1].item.Val)
			}
			yyVAL.item = yyDollar[1].item
		}
	case 70:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			unquoted := yylex.(*parser).unquoteString(yyDollar[1].item.Val)
			if !model.UTF8Validation.IsValidLabelName(unquoted) {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "invalid label name for grouping: %q", unquoted)
			}
			yyVAL.item = yyDollar[1].item
			yyVAL.item.Pos++
			yyVAL.item.Val = unquoted
		}
	case 71:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("grouping opts", "label")
			yyVAL.item = Item{}
		}
	case 72:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[2].node.(*NumberLiteral)
		}
	case 73:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			nl := yyDollar[3].node.(*NumberLiteral)
			if yyDollar[2].item.Typ == SUB {
				nl.Val *= -1
			}
			nl.PosRange.Start = yyDollar[2].item.Pos
			yyVAL.node = nl
		}
	case 74:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			fn, exist := getFunction(yyDollar[1].item.Val, yylex.(*parser).functions)
			if !exist {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "unknown function with name %q", yyDollar[1].item.Val)
			}
			if fn != nil && fn.Experimental && !yylex.(*parser).options.EnableExperimentalFunctions {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "function %q is not enabled", yyDollar[1].item.Val)
			}
			yyVAL.node = &Call{
				Func: fn,
				Args: yyDollar[2].node.(Expressions),
				PosRange: posrange.PositionRange{
					Start: yyDollar[1].item.Pos,
					End:   yylex.(*parser).lastClosing,
				},
			}
		}
	case 75:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			fn, exist := getFunction(yyDollar[1].item.Val, yylex.(*parser).functions)
			if !exist {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "unknown function with name %q", yyDollar[1].item.Val)
			}
			if fn != nil && fn.Experimental && !yylex.(*parser).options.EnableExperimentalFunctions {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "function %q is not enabled", yyDollar[1].item.Val)
			}
			yyVAL.node = &Call{
				Func: fn,
				Args: yyDollar[2].node.(Expressions),
				PosRange: posrange.PositionRange{
					Start: yyDollar[1].item.PositionRange().Start,
					End:   yylex.(*parser).lastClosing,
				},
			}
		}
	case 76:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			fn, exist := getFunction(yyDollar[1].item.Val, yylex.(*parser).functions)
			if !exist {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "unknown function with name %q", yyDollar[1].item.Val)
			}
			if fn != nil && fn.Experimental && !yylex.(*parser).options.EnableExperimentalFunctions {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "function %q is not enabled", yyDollar[1].item.Val)
			}
			yyVAL.node = &Call{
				Func: fn,
				Args: yyDollar[2].node.(Expressions),
				PosRange: posrange.PositionRange{
					Start: yyDollar[1].item.PositionRange().Start,
					End:   yylex.(*parser).lastClosing,
				},
			}
		}
	case 77:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			fn, exist := getFunction(yyDollar[1].item.Val, yylex.(*parser).functions)
			if !exist {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "unknown function with name %q", yyDollar[1].item.Val)
			}
			if fn != nil && fn.Experimental && !yylex.(*parser).options.EnableExperimentalFunctions {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "function %q is not enabled", yyDollar[1].item.Val)
			}
			yyVAL.node = &Call{
				Func: fn,
				Args: yyDollar[2].node.(Expressions),
				PosRange: posrange.PositionRange{
					Start: yyDollar[1].item.PositionRange().Start,
					End:   yylex.(*parser).lastClosing,
				},
			}
		}
	case 78:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = yyDollar[2].node
		}
	case 79:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.node = Expressions{}
		}
	case 80:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = append(yyDollar[1].node.(Expressions), yyDollar[3].node.(Expr))
		}
	case 81:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.node = Expressions{yyDollar[1].node.(Expr)}
		}
	case 82:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).addParseErrf(yyDollar[2].item.PositionRange(), "trailing commas not allowed in function call args")
			yyVAL.node = yyDollar[1].node
		}
	case 83:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &ParenExpr{Expr: yyDollar[2].node.(Expr), PosRange: mergeRanges(&yyDollar[1].item, &yyDollar[3].item)}
		}
	case 84:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			if numLit, ok := yyDollar[1].node.(*NumberLiteral); ok {
				if numLit.Val <= 0 {
					yylex.(*parser).addParseErrf(numLit.PositionRange(), "duration must be greater than 0")
					yyVAL.node = &NumberLiteral{Val: 0} // Return 0 on error.
					break
				}
				yyVAL.node = yyDollar[1].node
				break
			}
			yyVAL.node = yyDollar[1].node
		}
	case 85:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			if numLit, ok := yyDollar[3].node.(*NumberLiteral); ok {
				yylex.(*parser).addOffset(yyDollar[1].node, time.Duration(math.Round(numLit.Val*float64(time.Second))))
				yyVAL.node = yyDollar[1].node
				break
			}
			yylex.(*parser).addOffsetExpr(yyDollar[1].node, yyDollar[3].node.(*DurationExpr))
			yyVAL.node = yyDollar[1].node
		}
	case 86:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("offset", "number, duration, step(), or range()")
			yyVAL.node = yyDollar[1].node
		}
	case 87:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).setAnchored(yyDollar[1].node)
		}
	case 88:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).setSmoothed(yyDollar[1].node)
		}
	case 89:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).setTimestamp(yyDollar[1].node, yyDollar[3].float)
			yyVAL.node = yyDollar[1].node
		}
	case 90:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			yylex.(*parser).setAtModifierPreprocessor(yyDollar[1].node, yyDollar[3].item)
			yyVAL.node = yyDollar[1].node
		}
	case 91:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("@", "timestamp")
			yyVAL.node = yyDollar[1].node
		}
	case 94:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			var errMsg string
			vs, ok := yyDollar[1].node.(*VectorSelector)
			if !ok {
				errMsg = "ranges only allowed for vector selectors"
			} else if vs.OriginalOffset != 0 {
				errMsg = "no offset modifiers allowed before range"
			} else if vs.Timestamp != nil {
				errMsg = "no @ modifiers allowed before range"
			}

			if errMsg != "" {
				errRange := mergeRanges(&yyDollar[2].item, &yyDollar[4].item)
				yylex.(*parser).addParseErrf(errRange, "%s", errMsg)
			}

			var rangeNl time.Duration
			if numLit, ok := yyDollar[3].node.(*NumberLiteral); ok {
				rangeNl = time.Duration(math.Round(numLit.Val * float64(time.Second)))
			}
			rangeExpr, _ := yyDollar[3].node.(*DurationExpr)
			yyVAL.node = &MatrixSelector{
				VectorSelector: yyDollar[1].node.(Expr),
				Range:          rangeNl,
				RangeExpr:      rangeExpr,
				EndPos:         yylex.(*parser).lastClosing,
			}
		}
	case 95:
		yyDollar = yyS[yypt-6 : yypt+1]
		{
			var rangeNl time.Duration
			var stepNl time.Duration
			if numLit, ok := yyDollar[3].node.(*NumberLiteral); ok {
				rangeNl = time.Duration(math.Round(numLit.Val * float64(time.Second)))
			}
			rangeExpr, _ := yyDollar[3].node.(*DurationExpr)
			if numLit, ok := yyDollar[5].node.(*NumberLiteral); ok {
				stepNl = time.Duration(math.Round(numLit.Val * float64(time.Second)))
			}
			stepExpr, _ := yyDollar[5].node.(*DurationExpr)
			yyVAL.node = &SubqueryExpr{
				Expr:      yyDollar[1].node.(Expr),
				Range:     rangeNl,
				RangeExpr: rangeExpr,
				Step:      stepNl,
				StepExpr:  stepExpr,
				EndPos:    yyDollar[6].item.Pos + 1,
			}
		}
	case 96:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			var rangeNl time.Duration
			if numLit, ok := yyDollar[3].node.(*NumberLiteral); ok {
				rangeNl = time.Duration(math.Round(numLit.Val * float64(time.Second)))
			}
			rangeExpr, _ := yyDollar[3].node.(*DurationExpr)
			yyVAL.node = &SubqueryExpr{
				Expr:      yyDollar[1].node.(Expr),
				Range:     rangeNl,
				RangeExpr: rangeExpr,
				EndPos:    yyDollar[5].item.Pos + 1,
			}
		}
	case 97:
		yyDollar = yyS[yypt-6 : yypt+1]
		{
			yylex.(*parser).unexpected("subquery selector", "\"]\"")
			yyVAL.node = yyDollar[1].node
		}
	case 98:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			yylex.(*parser).unexpected("subquery selector", "number, duration, step(), range(), or \"]\"")
			yyVAL.node = yyDollar[1].node
		}
	case 99:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yylex.(*parser).unexpected("subquery or range", "\":\" or \"]\"")
			yyVAL.node = yyDollar[1].node
		}
	case 100:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("subquery or range selector", "number, duration, step(), or range()")
			yyVAL.node = yyDollar[1].node
		}
	case 101:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			if nl, ok := yyDollar[2].node.(*NumberLiteral); ok {
				if yyDollar[1].item.Typ == SUB {
					nl.Val *= -1
				}
				nl.PosRange.Start = yyDollar[1].item.Pos
				yyVAL.node = nl
			} else {
				yyVAL.node = &UnaryExpr{Op: yyDollar[1].item.Typ, Expr: yyDollar[2].node.(Expr), StartPos: yyDollar[1].item.Pos}
			}
		}
	case 102:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			vs := yyDollar[2].node.(*VectorSelector)
			vs.PosRange = mergeRanges(&yyDollar[1].item, vs)
			vs.Name = yyDollar[1].item.Val
			yylex.(*parser).assembleVectorSelector(vs)
			yyVAL.node = vs
		}
	case 103:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			vs := &VectorSelector{
				Name:          yyDollar[1].item.Val,
				LabelMatchers: []*labels.Matcher{},
				PosRange:      yyDollar[1].item.PositionRange(),
			}
			yylex.(*parser).assembleVectorSelector(vs)
			yyVAL.node = vs
		}
	case 104:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			vs := yyDollar[1].node.(*VectorSelector)
			yylex.(*parser).assembleVectorSelector(vs)
			yyVAL.node = vs
		}
	case 105:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &VectorSelector{
				LabelMatchers: yyDollar[2].matchers,
				PosRange:      mergeRanges(&yyDollar[1].item, &yyDollar[3].item),
			}
		}
	case 106:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = &VectorSelector{
				LabelMatchers: yyDollar[2].matchers,
				PosRange:      mergeRanges(&yyDollar[1].item, &yyDollar[4].item),
			}
		}
	case 107:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.node = &VectorSelector{
				LabelMatchers: []*labels.Matcher{},
				PosRange:      mergeRanges(&yyDollar[1].item, &yyDollar[2].item),
			}
		}
	case 108:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			if yyDollar[1].matchers != nil {
				yyVAL.matchers = append(yyDollar[1].matchers, yyDollar[3].matcher)
			} else {
				yyVAL.matchers = yyDollar[1].matchers
			}
		}
	case 109:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.matchers = []*labels.Matcher{yyDollar[1].matcher}
		}
	case 110:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("label matching", "\",\" or \"}\"")
			yyVAL.matchers = yyDollar[1].matchers
		}
	case 111:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.matcher = yylex.(*parser).newLabelMatcher(yyDollar[1].item, yyDollar[2].item, yyDollar[3].item)
		}
	case 112:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.matcher = yylex.(*parser).newLabelMatcher(yyDollar[1].item, yyDollar[2].item, yyDollar[3].item)
		}
	case 113:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.matcher = yylex.(*parser).newMetricNameMatcher(yyDollar[1].item)
		}
	case 114:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("label matching", "string")
			yyVAL.matcher = nil
		}
	case 115:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("label matching", "string")
			yyVAL.matcher = nil
		}
	case 116:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("label matching", "label matching operator")
			yyVAL.matcher = nil
		}
	case 117:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("label matching", "identifier or \"}\"")
			yyVAL.matcher = nil
		}
	case 118:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			b := labels.NewBuilder(yyDollar[2].labels)
			b.Set(labels.MetricName, yyDollar[1].item.Val)
			yyVAL.labels = b.Labels()
		}
	case 119:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.labels = yyDollar[1].labels
		}
	case 151:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.labels = labels.New(yyDollar[2].lblList...)
		}
	case 152:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.labels = labels.New(yyDollar[2].lblList...)
		}
	case 153:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.labels = labels.New()
		}
	case 154:
		yyDollar = yyS[yypt-0 : yypt+1]
		{
			yyVAL.labels = labels.New()
		}
	case 155:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.lblList = append(yyDollar[1].lblList, yyDollar[3].label)
		}
	case 156:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.lblList = []labels.Label{yyDollar[1].label}
		}
	case 157:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("label set", "\",\" or \"}\"")
			yyVAL.lblList = yyDollar[1].lblList
		}
	case 158:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.label = labels.Label{Name: yyDollar[1].item.Val, Value: yylex.(*parser).unquoteString(yyDollar[3].item.Val)}
		}
	case 159:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.label = labels.Label{Name: yyDollar[1].item.Val, Value: yylex.(*parser).unquoteString(yyDollar[3].item.Val)}
		}
	case 160:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.label = labels.Label{Name: labels.MetricName, Value: yyDollar[1].item.Val}
		}
	case 161:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("label set", "string")
			yyVAL.label = labels.Label{}
		}
	case 162:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).unexpected("label set", "string")
			yyVAL.label = labels.Label{}
		}
	case 163:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("label set", "\"=\"")
			yyVAL.label = labels.Label{}
		}
	case 164:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("label set", "identifier or \"}\"")
			yyVAL.label = labels.Label{}
		}
	case 165:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).generatedParserResult = &seriesDescription{
				labels: yyDollar[1].labels,
				values: yyDollar[2].series,
			}
		}
	case 166:
		yyDollar = yyS[yypt-0 : yypt+1]
		{
			yyVAL.series = []SequenceValue{}
		}
	case 167:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.series = append(yyDollar[1].series, yyDollar[3].series...)
		}
	case 168:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.series = yyDollar[1].series
		}
	case 169:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yylex.(*parser).unexpected("series values", "")
			yyVAL.series = nil
		}
	case 170:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.series = []SequenceValue{{Omitted: true}}
		}
	case 171:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.series = []SequenceValue{}
			for i := uint64(0); i < yyDollar[3].uint; i++ {
				yyVAL.series = append(yyVAL.series, SequenceValue{Omitted: true})
			}
		}
	case 172:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.series = []SequenceValue{{Value: yyDollar[1].float}}
		}
	case 173:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.series = []SequenceValue{}
			// Add an additional value for time 0, which we ignore in tests.
			for i := uint64(0); i <= yyDollar[3].uint; i++ {
				yyVAL.series = append(yyVAL.series, SequenceValue{Value: yyDollar[1].float})
			}
		}
	case 174:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.series = []SequenceValue{}
			// Add an additional value for time 0, which we ignore in tests.
			for i := uint64(0); i <= yyDollar[4].uint; i++ {
				yyVAL.series = append(yyVAL.series, SequenceValue{Value: yyDollar[1].float})
				yyDollar[1].float += yyDollar[2].float
			}
		}
	case 175:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.series = []SequenceValue{yylex.(*parser).newHistogramSequenceValue(yyDollar[1].histogram)}
		}
	case 176:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.series = []SequenceValue{}
			// Add an additional value for time 0, which we ignore in tests.
			sv := yylex.(*parser).newHistogramSequenceValue(yyDollar[1].histogram)
			for i := uint64(0); i <= yyDollar[3].uint; i++ {
				yyVAL.series = append(yyVAL.series, sv)
				//$1 += $2
			}
		}
	case 177:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			val, err := yylex.(*parser).histogramsIncreaseSeries(yyDollar[1].histogram, yyDollar[3].histogram, yyDollar[5].uint)
			if err != nil {
				yylex.(*parser).addSemanticError(err)
			}
			yyVAL.series = val
		}
	case 178:
		yyDollar = yyS[yypt-5 : yypt+1]
		{
			val, err := yylex.(*parser).histogramsDecreaseSeries(yyDollar[1].histogram, yyDollar[3].histogram, yyDollar[5].uint)
			if err != nil {
				yylex.(*parser).addSemanticError(err)
			}
			yyVAL.series = val
		}
	case 179:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			if yyDollar[1].item.Val != "stale" {
				yylex.(*parser).unexpected("series values", "number or \"stale\"")
			}
			yyVAL.float = math.Float64frombits(value.StaleNaN)
		}
	case 182:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.histogram = yylex.(*parser).buildHistogramFromMap(&yyDollar[2].descriptors)
		}
	case 183:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.histogram = yylex.(*parser).buildHistogramFromMap(&yyDollar[2].descriptors)
		}
	case 184:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			m := yylex.(*parser).newMap()
			yyVAL.histogram = yylex.(*parser).buildHistogramFromMap(&m)
		}
	case 185:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			m := yylex.(*parser).newMap()
			yyVAL.histogram = yylex.(*parser).buildHistogramFromMap(&m)
		}
	case 186:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = *(yylex.(*parser).mergeMaps(&yyDollar[1].descriptors, &yyDollar[3].descriptors))
		}
	case 187:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.descriptors = yyDollar[1].descriptors
		}
	case 188:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yylex.(*parser).unexpected("histogram description", "histogram description key, e.g. buckets:[5 10 7]")
		}
	case 189:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["schema"] = yyDollar[3].int
		}
	case 190:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["sum"] = yyDollar[3].float
		}
	case 191:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["count"] = yyDollar[3].float
		}
	case 192:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["z_bucket"] = yyDollar[3].float
		}
	case 193:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["z_bucket_w"] = yyDollar[3].float
		}
	case 194:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["custom_values"] = yyDollar[3].bucket_set
		}
	case 195:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["buckets"] = yyDollar[3].bucket_set
		}
	case 196:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["offset"] = yyDollar[3].int
		}
	case 197:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["n_buckets"] = yyDollar[3].bucket_set
		}
	case 198:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["n_offset"] = yyDollar[3].int
		}
	case 199:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.descriptors = yylex.(*parser).newMap()
			yyVAL.descriptors["counter_reset_hint"] = yyDollar[3].item
		}
	case 200:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.bucket_set = yyDollar[2].bucket_set
		}
	case 201:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.bucket_set = yyDollar[2].bucket_set
		}
	case 202:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.bucket_set = append(yyDollar[1].bucket_set, yyDollar[3].float)
		}
	case 203:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.bucket_set = []float64{yyDollar[1].float}
		}
	case 265:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.node = &NumberLiteral{
				Val:      yylex.(*parser).number(yyDollar[1].item.Val),
				PosRange: yyDollar[1].item.PositionRange(),
			}
		}
	case 266:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			var err error
			var dur time.Duration
			dur, err = parseDuration(yyDollar[1].item.Val)
			if err != nil {
				yylex.(*parser).addParseErr(yyDollar[1].item.PositionRange(), err)
			}
			yyVAL.node = &NumberLiteral{
				Val:      dur.Seconds(),
				PosRange: yyDollar[1].item.PositionRange(),
				Duration: true,
			}
		}
	case 267:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.float = yylex.(*parser).number(yyDollar[1].item.Val)
		}
	case 268:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			var err error
			var dur time.Duration
			dur, err = parseDuration(yyDollar[1].item.Val)
			if err != nil {
				yylex.(*parser).addParseErr(yyDollar[1].item.PositionRange(), err)
			}
			yyVAL.float = dur.Seconds()
		}
	case 269:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.float = yyDollar[2].float
		}
	case 270:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.float = -yyDollar[2].float
		}
	case 273:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			var err error
			yyVAL.uint, err = strconv.ParseUint(yyDollar[1].item.Val, 10, 64)
			if err != nil {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "invalid repetition in series values: %s", err)
			}
		}
	case 274:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			yyVAL.int = -int64(yyDollar[2].uint)
		}
	case 275:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.int = int64(yyDollar[1].uint)
		}
	case 276:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.node = &StringLiteral{
				Val:      yylex.(*parser).unquoteString(yyDollar[1].item.Val),
				PosRange: yyDollar[1].item.PositionRange(),
			}
		}
	case 277:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			yyVAL.item = Item{
				Typ: METRIC_IDENTIFIER,
				Pos: yyDollar[1].item.PositionRange().Start,
				Val: yylex.(*parser).unquoteString(yyDollar[1].item.Val),
			}
		}
	case 278:
		yyDollar = yyS[yypt-0 : yypt+1]
		{
			yyVAL.strings = nil
		}
	case 280:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			nl := yyDollar[1].node.(*NumberLiteral)
			if nl.Val > 1<<63/1e9 || nl.Val < -(1<<63)/1e9 {
				yylex.(*parser).addParseErrf(nl.PosRange, "duration out of range")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
			yyVAL.node = nl
		}
	case 281:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			nl := yyDollar[2].node.(*NumberLiteral)
			if yyDollar[1].item.Typ == SUB {
				nl.Val *= -1
			}
			if nl.Val > 1<<63/1e9 || nl.Val < -(1<<63)/1e9 {
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "duration out of range")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
			nl.PosRange.Start = yyDollar[1].item.Pos
			yyVAL.node = nl
		}
	case 282:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       STEP,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[3].item.PositionRange().End,
			}
		}
	case 283:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       RANGE,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[3].item.PositionRange().End,
			}
		}
	case 284:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op: yyDollar[1].item.Typ,
				RHS: &DurationExpr{
					Op:       STEP,
					StartPos: yyDollar[2].item.PositionRange().Start,
					EndPos:   yyDollar[4].item.PositionRange().End,
				},
				StartPos: yyDollar[1].item.Pos,
			}
		}
	case 285:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op: yyDollar[1].item.Typ,
				RHS: &DurationExpr{
					Op:       RANGE,
					StartPos: yyDollar[2].item.PositionRange().Start,
					EndPos:   yyDollar[4].item.PositionRange().End,
				},
				StartPos: yyDollar[1].item.Pos,
			}
		}
	case 286:
		yyDollar = yyS[yypt-6 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       yyDollar[1].item.Typ,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[6].item.PositionRange().End,
				LHS:      yyDollar[3].node.(Expr),
				RHS:      yyDollar[5].node.(Expr),
			}
		}
	case 287:
		yyDollar = yyS[yypt-7 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       yyDollar[1].item.Typ,
				StartPos: yyDollar[1].item.Pos,
				EndPos:   yyDollar[6].node.PositionRange().End,
				RHS: &DurationExpr{
					Op:       yyDollar[2].item.Typ,
					StartPos: yyDollar[2].item.PositionRange().Start,
					EndPos:   yyDollar[6].node.PositionRange().End,
					LHS:      yyDollar[4].node.(Expr),
					RHS:      yyDollar[6].node.(Expr),
				},
			}
		}
	case 288:
		yyDollar = yyS[yypt-4 : yypt+1]
		{
			de := yyDollar[3].node.(*DurationExpr)
			de.Wrapped = true
			if yyDollar[1].item.Typ == SUB {
				yyVAL.node = &DurationExpr{
					Op:       SUB,
					RHS:      de,
					StartPos: yyDollar[1].item.Pos,
				}
				break
			}
			yyVAL.node = yyDollar[3].node
		}
	case 292:
		yyDollar = yyS[yypt-1 : yypt+1]
		{
			nl := yyDollar[1].node.(*NumberLiteral)
			if nl.Val > 1<<63/1e9 || nl.Val < -(1<<63)/1e9 {
				yylex.(*parser).addParseErrf(nl.PosRange, "duration out of range")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
			yyVAL.node = nl
		}
	case 293:
		yyDollar = yyS[yypt-2 : yypt+1]
		{
			switch expr := yyDollar[2].node.(type) {
			case *NumberLiteral:
				if yyDollar[1].item.Typ == SUB {
					expr.Val *= -1
				}
				if expr.Val > 1<<63/1e9 || expr.Val < -(1<<63)/1e9 {
					yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "duration out of range")
					yyVAL.node = &NumberLiteral{Val: 0}
					break
				}
				expr.PosRange.Start = yyDollar[1].item.Pos
				yyVAL.node = expr
				break
			case *DurationExpr:
				if yyDollar[1].item.Typ == SUB {
					yyVAL.node = &DurationExpr{
						Op:       SUB,
						RHS:      expr,
						StartPos: yyDollar[1].item.Pos,
					}
					break
				}
				yyVAL.node = expr
				break
			default:
				yylex.(*parser).addParseErrf(yyDollar[1].item.PositionRange(), "expected number literal or duration expression")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
		}
	case 294:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			yyVAL.node = &DurationExpr{Op: ADD, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 295:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			yyVAL.node = &DurationExpr{Op: SUB, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 296:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			yyVAL.node = &DurationExpr{Op: MUL, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 297:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			if nl, ok := yyDollar[3].node.(*NumberLiteral); ok && nl.Val == 0 {
				yylex.(*parser).addParseErrf(yyDollar[2].item.PositionRange(), "division by zero")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
			yyVAL.node = &DurationExpr{Op: DIV, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 298:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			if nl, ok := yyDollar[3].node.(*NumberLiteral); ok && nl.Val == 0 {
				yylex.(*parser).addParseErrf(yyDollar[2].item.PositionRange(), "modulo by zero")
				yyVAL.node = &NumberLiteral{Val: 0}
				break
			}
			yyVAL.node = &DurationExpr{Op: MOD, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 299:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[1].node.(Expr))
			yyVAL.node = &DurationExpr{Op: POW, LHS: yyDollar[1].node.(Expr), RHS: yyDollar[3].node.(Expr)}
		}
	case 300:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       STEP,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[3].item.PositionRange().End,
			}
		}
	case 301:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       RANGE,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[3].item.PositionRange().End,
			}
		}
	case 302:
		yyDollar = yyS[yypt-6 : yypt+1]
		{
			yyVAL.node = &DurationExpr{
				Op:       yyDollar[1].item.Typ,
				StartPos: yyDollar[1].item.PositionRange().Start,
				EndPos:   yyDollar[6].item.PositionRange().End,
				LHS:      yyDollar[3].node.(Expr),
				RHS:      yyDollar[5].node.(Expr),
			}
		}
	case 304:
		yyDollar = yyS[yypt-3 : yypt+1]
		{
			yylex.(*parser).experimentalDurationExpr(yyDollar[2].node.(Expr))
			if durationExpr, ok := yyDollar[2].node.(*DurationExpr); ok {
				durationExpr.Wrapped = true
				yyVAL.node = durationExpr
				break
			}
			yyVAL.node = yyDollar[2].node
		}
	}
	goto yystack /* stack new state and value */
}
