// Code generated by protoc-gen-gogo. DO NOT EDIT.
// source: node.proto

package remoteexec

import (
	fmt "fmt"
	_ "github.com/gogo/protobuf/gogoproto"
	proto "github.com/gogo/protobuf/proto"
	io "io"
	math "math"
	math_bits "math/bits"
	reflect "reflect"
	strings "strings"
)

// Reference imports to suppress errors if they are not otherwise used.
var _ = proto.Marshal
var _ = fmt.Errorf
var _ = math.Inf

// This is a compile-time assertion to ensure that this generated file
// is compatible with the proto package it is being compiled against.
// A compilation error at this line likely means your copy of the
// proto package needs to be updated.
const _ = proto.GoGoProtoPackageIsVersion3 // please upgrade the proto package

type RemoteExecutionGroupDetails struct {
	// If the query has been sharded, we need to eagerly load and buffer the response from queriers in query-frontends
	// to avoid a possible deadlock: we need the series metadata for each sharded leg before we can start streaming series
	// data back to the caller. If we don't eagerly load responses from queriers, then in a query with (for example) 8
	// sharded legs, we might end up deadlocked waiting for some or all of the last legs because these requests can't
	// start either due to the query-frontend's per-query max parallelism limit or the number of available querier
	// workers. If we eagerly load responses, then exhausting the responses that have started will eventually make space
	// for the outstanding requests to complete.
	EagerLoad bool `protobuf:"varint,1,opt,name=eagerLoad,proto3" json:"eagerLoad,omitempty"`
}

func (m *RemoteExecutionGroupDetails) Reset()      { *m = RemoteExecutionGroupDetails{} }
func (*RemoteExecutionGroupDetails) ProtoMessage() {}
func (*RemoteExecutionGroupDetails) Descriptor() ([]byte, []int) {
	return fileDescriptor_0c843d59d2d938e7, []int{0}
}
func (m *RemoteExecutionGroupDetails) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *RemoteExecutionGroupDetails) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_RemoteExecutionGroupDetails.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *RemoteExecutionGroupDetails) XXX_Merge(src proto.Message) {
	xxx_messageInfo_RemoteExecutionGroupDetails.Merge(m, src)
}
func (m *RemoteExecutionGroupDetails) XXX_Size() int {
	return m.Size()
}
func (m *RemoteExecutionGroupDetails) XXX_DiscardUnknown() {
	xxx_messageInfo_RemoteExecutionGroupDetails.DiscardUnknown(m)
}

var xxx_messageInfo_RemoteExecutionGroupDetails proto.InternalMessageInfo

func (m *RemoteExecutionGroupDetails) GetEagerLoad() bool {
	if m != nil {
		return m.EagerLoad
	}
	return false
}

func (*RemoteExecutionGroupDetails) XXX_MessageName() string {
	return "remoteexec.RemoteExecutionGroupDetails"
}

type RemoteExecutionConsumerDetails struct {
	// Index of node on child RemoteExecutionGroup node that this consumer consumes from.
	NodeIndex uint64 `protobuf:"varint,1,opt,name=nodeIndex,proto3" json:"nodeIndex,omitempty"`
}

func (m *RemoteExecutionConsumerDetails) Reset()      { *m = RemoteExecutionConsumerDetails{} }
func (*RemoteExecutionConsumerDetails) ProtoMessage() {}
func (*RemoteExecutionConsumerDetails) Descriptor() ([]byte, []int) {
	return fileDescriptor_0c843d59d2d938e7, []int{1}
}
func (m *RemoteExecutionConsumerDetails) XXX_Unmarshal(b []byte) error {
	return m.Unmarshal(b)
}
func (m *RemoteExecutionConsumerDetails) XXX_Marshal(b []byte, deterministic bool) ([]byte, error) {
	if deterministic {
		return xxx_messageInfo_RemoteExecutionConsumerDetails.Marshal(b, m, deterministic)
	} else {
		b = b[:cap(b)]
		n, err := m.MarshalToSizedBuffer(b)
		if err != nil {
			return nil, err
		}
		return b[:n], nil
	}
}
func (m *RemoteExecutionConsumerDetails) XXX_Merge(src proto.Message) {
	xxx_messageInfo_RemoteExecutionConsumerDetails.Merge(m, src)
}
func (m *RemoteExecutionConsumerDetails) XXX_Size() int {
	return m.Size()
}
func (m *RemoteExecutionConsumerDetails) XXX_DiscardUnknown() {
	xxx_messageInfo_RemoteExecutionConsumerDetails.DiscardUnknown(m)
}

var xxx_messageInfo_RemoteExecutionConsumerDetails proto.InternalMessageInfo

func (m *RemoteExecutionConsumerDetails) GetNodeIndex() uint64 {
	if m != nil {
		return m.NodeIndex
	}
	return 0
}

func (*RemoteExecutionConsumerDetails) XXX_MessageName() string {
	return "remoteexec.RemoteExecutionConsumerDetails"
}
func init() {
	proto.RegisterType((*RemoteExecutionGroupDetails)(nil), "remoteexec.RemoteExecutionGroupDetails")
	proto.RegisterType((*RemoteExecutionConsumerDetails)(nil), "remoteexec.RemoteExecutionConsumerDetails")
}

func init() { proto.RegisterFile("node.proto", fileDescriptor_0c843d59d2d938e7) }

var fileDescriptor_0c843d59d2d938e7 = []byte{
	// 207 bytes of a gzipped FileDescriptorProto
	0x1f, 0x8b, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0xff, 0xe2, 0xe2, 0xca, 0xcb, 0x4f, 0x49,
	0xd5, 0x2b, 0x28, 0xca, 0x2f, 0xc9, 0x17, 0xe2, 0x2a, 0x4a, 0xcd, 0xcd, 0x2f, 0x49, 0x4d, 0xad,
	0x48, 0x4d, 0x96, 0x12, 0x49, 0xcf, 0x4f, 0xcf, 0x07, 0x0b, 0xeb, 0x83, 0x58, 0x10, 0x15, 0x4a,
	0xd6, 0x5c, 0xd2, 0x41, 0x60, 0x35, 0xae, 0x15, 0xa9, 0xc9, 0xa5, 0x25, 0x99, 0xf9, 0x79, 0xee,
	0x45, 0xf9, 0xa5, 0x05, 0x2e, 0xa9, 0x25, 0x89, 0x99, 0x39, 0xc5, 0x42, 0x32, 0x5c, 0x9c, 0xa9,
	0x89, 0xe9, 0xa9, 0x45, 0x3e, 0xf9, 0x89, 0x29, 0x12, 0x8c, 0x0a, 0x8c, 0x1a, 0x1c, 0x41, 0x08,
	0x01, 0x25, 0x3b, 0x2e, 0x39, 0x34, 0xcd, 0xce, 0xf9, 0x79, 0xc5, 0xa5, 0xb9, 0xa9, 0x45, 0x48,
	0xfa, 0x41, 0xce, 0xf1, 0xcc, 0x4b, 0x49, 0xad, 0x00, 0xeb, 0x67, 0x09, 0x42, 0x08, 0x38, 0x59,
	0x5c, 0x78, 0x28, 0xc7, 0x70, 0xe3, 0xa1, 0x1c, 0xc3, 0x87, 0x87, 0x72, 0x8c, 0x0d, 0x8f, 0xe4,
	0x18, 0x57, 0x3c, 0x92, 0x63, 0x38, 0xf1, 0x48, 0x8e, 0xf1, 0xc2, 0x23, 0x39, 0xc6, 0x07, 0x8f,
	0xe4, 0x18, 0x5f, 0x3c, 0x92, 0x63, 0xf8, 0xf0, 0x48, 0x8e, 0x71, 0xc2, 0x63, 0x39, 0x86, 0x13,
	0x8f, 0xe5, 0x18, 0x2f, 0x3c, 0x96, 0x63, 0xb8, 0xf1, 0x58, 0x8e, 0x21, 0x89, 0x0d, 0xec, 0x7a,
	0x63, 0x40, 0x00, 0x00, 0x00, 0xff, 0xff, 0xd2, 0x9a, 0xe2, 0x11, 0xed, 0x00, 0x00, 0x00,
}

func (this *RemoteExecutionGroupDetails) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 5)
	s = append(s, "&remoteexec.RemoteExecutionGroupDetails{")
	s = append(s, "EagerLoad: "+fmt.Sprintf("%#v", this.EagerLoad)+",\n")
	s = append(s, "}")
	return strings.Join(s, "")
}
func (this *RemoteExecutionConsumerDetails) GoString() string {
	if this == nil {
		return "nil"
	}
	s := make([]string, 0, 5)
	s = append(s, "&remoteexec.RemoteExecutionConsumerDetails{")
	s = append(s, "NodeIndex: "+fmt.Sprintf("%#v", this.NodeIndex)+",\n")
	s = append(s, "}")
	return strings.Join(s, "")
}
func valueToGoStringNode(v interface{}, typ string) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("func(v %v) *%v { return &v } ( %#v )", typ, typ, pv)
}
func (m *RemoteExecutionGroupDetails) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *RemoteExecutionGroupDetails) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *RemoteExecutionGroupDetails) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if m.EagerLoad {
		i--
		if m.EagerLoad {
			dAtA[i] = 1
		} else {
			dAtA[i] = 0
		}
		i--
		dAtA[i] = 0x8
	}
	return len(dAtA) - i, nil
}

func (m *RemoteExecutionConsumerDetails) Marshal() (dAtA []byte, err error) {
	size := m.Size()
	dAtA = make([]byte, size)
	n, err := m.MarshalToSizedBuffer(dAtA[:size])
	if err != nil {
		return nil, err
	}
	return dAtA[:n], nil
}

func (m *RemoteExecutionConsumerDetails) MarshalTo(dAtA []byte) (int, error) {
	size := m.Size()
	return m.MarshalToSizedBuffer(dAtA[:size])
}

func (m *RemoteExecutionConsumerDetails) MarshalToSizedBuffer(dAtA []byte) (int, error) {
	i := len(dAtA)
	_ = i
	var l int
	_ = l
	if m.NodeIndex != 0 {
		i = encodeVarintNode(dAtA, i, uint64(m.NodeIndex))
		i--
		dAtA[i] = 0x8
	}
	return len(dAtA) - i, nil
}

func encodeVarintNode(dAtA []byte, offset int, v uint64) int {
	offset -= sovNode(v)
	base := offset
	for v >= 1<<7 {
		dAtA[offset] = uint8(v&0x7f | 0x80)
		v >>= 7
		offset++
	}
	dAtA[offset] = uint8(v)
	return base
}
func (m *RemoteExecutionGroupDetails) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if m.EagerLoad {
		n += 2
	}
	return n
}

func (m *RemoteExecutionConsumerDetails) Size() (n int) {
	if m == nil {
		return 0
	}
	var l int
	_ = l
	if m.NodeIndex != 0 {
		n += 1 + sovNode(uint64(m.NodeIndex))
	}
	return n
}

func sovNode(x uint64) (n int) {
	return (math_bits.Len64(x|1) + 6) / 7
}
func sozNode(x uint64) (n int) {
	return sovNode(uint64((x << 1) ^ uint64((int64(x) >> 63))))
}
func (this *RemoteExecutionGroupDetails) String() string {
	if this == nil {
		return "nil"
	}
	s := strings.Join([]string{`&RemoteExecutionGroupDetails{`,
		`EagerLoad:` + fmt.Sprintf("%v", this.EagerLoad) + `,`,
		`}`,
	}, "")
	return s
}
func (this *RemoteExecutionConsumerDetails) String() string {
	if this == nil {
		return "nil"
	}
	s := strings.Join([]string{`&RemoteExecutionConsumerDetails{`,
		`NodeIndex:` + fmt.Sprintf("%v", this.NodeIndex) + `,`,
		`}`,
	}, "")
	return s
}
func valueToStringNode(v interface{}) string {
	rv := reflect.ValueOf(v)
	if rv.IsNil() {
		return "nil"
	}
	pv := reflect.Indirect(rv).Interface()
	return fmt.Sprintf("*%v", pv)
}
func (m *RemoteExecutionGroupDetails) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowNode
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: RemoteExecutionGroupDetails: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: RemoteExecutionGroupDetails: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field EagerLoad", wireType)
			}
			var v int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				v |= int(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			m.EagerLoad = bool(v != 0)
		default:
			iNdEx = preIndex
			skippy, err := skipNode(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if (skippy < 0) || (iNdEx+skippy) < 0 {
				return ErrInvalidLengthNode
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func (m *RemoteExecutionConsumerDetails) Unmarshal(dAtA []byte) error {
	l := len(dAtA)
	iNdEx := 0
	for iNdEx < l {
		preIndex := iNdEx
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return ErrIntOverflowNode
			}
			if iNdEx >= l {
				return io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= uint64(b&0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		fieldNum := int32(wire >> 3)
		wireType := int(wire & 0x7)
		if wireType == 4 {
			return fmt.Errorf("proto: RemoteExecutionConsumerDetails: wiretype end group for non-group")
		}
		if fieldNum <= 0 {
			return fmt.Errorf("proto: RemoteExecutionConsumerDetails: illegal tag %d (wire type %d)", fieldNum, wire)
		}
		switch fieldNum {
		case 1:
			if wireType != 0 {
				return fmt.Errorf("proto: wrong wireType = %d for field NodeIndex", wireType)
			}
			m.NodeIndex = 0
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return ErrIntOverflowNode
				}
				if iNdEx >= l {
					return io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				m.NodeIndex |= uint64(b&0x7F) << shift
				if b < 0x80 {
					break
				}
			}
		default:
			iNdEx = preIndex
			skippy, err := skipNode(dAtA[iNdEx:])
			if err != nil {
				return err
			}
			if (skippy < 0) || (iNdEx+skippy) < 0 {
				return ErrInvalidLengthNode
			}
			if (iNdEx + skippy) > l {
				return io.ErrUnexpectedEOF
			}
			iNdEx += skippy
		}
	}

	if iNdEx > l {
		return io.ErrUnexpectedEOF
	}
	return nil
}
func skipNode(dAtA []byte) (n int, err error) {
	l := len(dAtA)
	iNdEx := 0
	depth := 0
	for iNdEx < l {
		var wire uint64
		for shift := uint(0); ; shift += 7 {
			if shift >= 64 {
				return 0, ErrIntOverflowNode
			}
			if iNdEx >= l {
				return 0, io.ErrUnexpectedEOF
			}
			b := dAtA[iNdEx]
			iNdEx++
			wire |= (uint64(b) & 0x7F) << shift
			if b < 0x80 {
				break
			}
		}
		wireType := int(wire & 0x7)
		switch wireType {
		case 0:
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowNode
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				iNdEx++
				if dAtA[iNdEx-1] < 0x80 {
					break
				}
			}
		case 1:
			iNdEx += 8
		case 2:
			var length int
			for shift := uint(0); ; shift += 7 {
				if shift >= 64 {
					return 0, ErrIntOverflowNode
				}
				if iNdEx >= l {
					return 0, io.ErrUnexpectedEOF
				}
				b := dAtA[iNdEx]
				iNdEx++
				length |= (int(b) & 0x7F) << shift
				if b < 0x80 {
					break
				}
			}
			if length < 0 {
				return 0, ErrInvalidLengthNode
			}
			iNdEx += length
		case 3:
			depth++
		case 4:
			if depth == 0 {
				return 0, ErrUnexpectedEndOfGroupNode
			}
			depth--
		case 5:
			iNdEx += 4
		default:
			return 0, fmt.Errorf("proto: illegal wireType %d", wireType)
		}
		if iNdEx < 0 {
			return 0, ErrInvalidLengthNode
		}
		if depth == 0 {
			return iNdEx, nil
		}
	}
	return 0, io.ErrUnexpectedEOF
}

var (
	ErrInvalidLengthNode        = fmt.Errorf("proto: negative length found during unmarshaling")
	ErrIntOverflowNode          = fmt.Errorf("proto: integer overflow")
	ErrUnexpectedEndOfGroupNode = fmt.Errorf("proto: unexpected end of group")
)
